import logging
import os

from multiprocessing import cpu_count


from mercury_agent.capabilities import capability
from mercury.common.helpers import cli


log = logging.getLogger(__name__)


class Stress(object):
    """
    Class for running our stress commands.
    """

    def __init__(self, stress_path="/usr/bin/stress"):
        """
        Make sure the stress binary is found
        """
        self.stress_path = stress_path
        if not os.path.exists(stress_path):
            raise Exception('Stress binary not found at '
                            '{}.'.format(self.stress_path))

    @staticmethod
    def get_system_memory():
        """
        Gets the amount of system memory in a server.
        """
        log.info('Getting the system memory')
        cmd = 'grep MemFree /proc/meminfo'
        result = cli.run(cmd)

        if not result.returncode:
            k = result.stdout.split()[1].strip()
            g = int(k)/1024/1024
            return '{}G'.format(int(g))

    def killall(self):
        """
        The way the stress binary behaves makes it hard to kill
        """
        log.error('Killing Stress Process.')
        os.system('killall stress')

    def cpu_stress(self, timeout):
        """
        Stress test the cpu on all cores.
        """
        log.info('Stress testing the cpu for %s seconds.' % timeout)
        cmd = '{0} --cpu {1} -t {2}'.format(self.stress_path,
                                            cpu_count(),
                                            timeout)
        log.info('Executing: {}'.format(cmd))
        result = cli.run(cmd)
        if result.returncode:
            self.killall()

    def memory_stress(self, timeout, amount=None):
        """
        Stress test the memory half of free memory.

        Amount should be in 1024K, 1024M, 1024G format.
        """

        # If we did not pass in a specific memory lets
        # try and run against all system memory, but if that
        # fails we default to 1024M.
        if not amount:
            amount = self.get_system_memory()
            if not amount:
                amount = '1024M'

        log.info('Stressing memory for %s seconds at %s.' %
                         (timeout, amount))
        cmd = '{0} --vm 1 --vm-bytes {1} --vm-hang 5 -t {2}'.format(
            self.stress_path, amount, timeout)
        log.info('Executing: {}'.format(cmd))
        result = cli.run(cmd)
        if result.returncode:
            self.killall()


@capability('memory_stress_test',
            description='Run stress on memory for specified amount of time',
            num_args=1, serial=False, timeout=3600)
def memory_stress_test(stress_seconds):
    """

    :param stress_seconds:
    :return:
    """
    log.info('Starting stress test on CPU and Memory for '
             '{} seconds'.format(stress_seconds))
    stress_tester = Stress()
    stress_tester.memory_stress(stress_seconds)


@capability('cpu_stress_test',
            description='Run stress on cpu for specified amount of time',
            num_args=1, serial=False, timeout=3600)
def cpu_stress_test(stress_seconds):
    """

    :param stress_seconds:
    :return:
    """
    log.info('Starting stress test on CPU and Memory for '
             '{} seconds'.format(stress_seconds))
    stress_tester = Stress()
    stress_tester.cpu_stress(stress_seconds)



# class BMCMethod(object):
#     name = ''
#
#     def __init__(self, handler):
#         self.handler = handler
#
#     def __call__(self, *args, **kwargs):
#         method = getattr(self.handler, self.name)
#         return method(*args, **kwargs)
#
#
#
# class OBMReset(BMCMethod):
#     name = 'reset_hard'
#
#
#
# obm_reset = OBMReset(handler)
# add_capability
# handler = get_bmc_handler(global_device_info['dmi']['sys_vendor'])



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
The current device routing table
"""

from . import inspector
from mercury_agent.inspector.hwlib.iproute2 import IPRoute2


@inspector.expose('routes')
def route_inspector():
    return IPRoute2().table


def find_default_route(routes):
    """

    :param routes:
    :return:
    """
    _defaults = []
    for route in routes:
        if route.get('destination') == 'default':
            if 'metric' not in route:
                route['metric'] = 0
            _defaults.append(route)

    if not _defaults:
        return {}

    return sorted(_defaults, key=lambda _d: int(_d['metric']))[0]


if __name__ == '__main__':
    from pprint import pprint
    table = route_inspector()
    pprint(table)
    pprint('-' * 10)
    pprint(find_default_route(table))



from io import BytesIO
from lxml import etree


class XMLError(Exception):
    pass


class XLoader(object):
    def __init__(self, xml_data):
        self.xml_data = xml_data

    @property
    def root(self):
        return etree.parse(BytesIO(self.xml_data)).getroot()


class XMLAbout(dict):
    def __init__(self, oma):
        super(XMLAbout, self).__init__()
        about = oma.find('About')
        if about is None:
            raise XMLError('About element is missing.')

        self['components'] = list()
        for child in about:
            if child.tag == 'Component':
                self._add_component(child)
            else:
                self[child.tag] = child.text.strip()

    def _add_component(self, component):
        d = dict()
        for child in component:
            d[child.tag] = child.text.strip()
        self['components'].append(d)


class XMLChassisStatus(dict):
    status_translation = {
        1: "Unknown",
        2: "Ok",
        3: "Non-critical",
        4: "Critical"
    }

    # triple (tag, object list tag, description)
    element_list = [
        ('intrusion', 'IntrusionObj', 'IntrusionLoc'),
        ('voltages', 'VoltageObj', 'ProbeLocation'),
        ('temperatures', 'TemperatureObj', 'ProbeLocation'),
        ('fans', 'Redundancy', 'Fan System'),
        ('currents', 'CurrentObj', 'ProbeLocation'),
        ('powersupply', 'PowerSupplyObj', 'PSLocation'),
        ('powermonitoring', 'PowerConsumptionDataObj', 'Identifier'),
        ('processor', 'DevProcessorObj', 'Brand'),
        ('esmlog', 'LogObj', 'HardwareLog'),
        ('memory', 'MemDevObj', 'Memory'),
        ('batteries', 'BatteryObj', 'ProbeLocation'),
        ('sdcard', 'SDCard', 'SDCardLocation')
    ]

    computed_status_tag = 'computedobjstatus'
    objstatus_tag = 'objstatus'

    def __init__(self, root):
        super(XMLChassisStatus, self).__init__()
        parent = root[0]
        for tag, obj, obj_desc in self.element_list:
            element = parent.find(tag)
            if element is None:
                continue
            self[tag] = dict()
            status_element = element.find(self.computed_status_tag)
            self[tag]['status'] = int(status_element.text.strip())
            self[tag]['status_string'] = status_element.attrib.get('strval')
            self[tag]['sensors'] = list()

            sensor_elements = element.findall(obj)
            for sensor_element in sensor_elements:
                sensor_dict = dict()
                description_element = sensor_element.find(obj_desc)
                if description_element is not None:
                    sensor_dict['description'] = description_element.text
                else:
                    sensor_dict['description'] = obj_desc

                sensor_dict['status'] = sensor_element.find(self.objstatus_tag).text
                self[tag]['sensors'].append(sensor_dict)

    @property
    def errors(self):
        err = list()
        for tag in self:
            if self[tag]['status'] == 2:
                continue
            error_dict = self[tag]
            error_dict['component'] = tag
            err.append(error_dict)

        return err


class XMLVDisk(dict):
    vdisk_obj_tag = 'DCStorageObject'

    def __init__(self, oma):
        super(XMLVDisk, self).__init__()
        vdisk_element = oma.find('VirtualDisks')
        if vdisk_element is None:
            return

        obj_elements = vdisk_element.findall(self.vdisk_obj_tag)
        for obj in obj_elements:
            device_id = int(obj.find('DeviceID').text)
            self[device_id] = dict()
            self[device_id]['status'] = int(obj.find('ObjStatus').text)
            self[device_id]['state'] = int(obj.find('ObjState').text)
            self[device_id]['read_policy'] = int(obj.find('DefaultReadPolicy').text)
            self[device_id]['write_policy'] = int(obj.find('DefaultWritePolicy').text)
            self[device_id]['name'] = obj.find('Name').text


class XMLController(dict):
    controller_obj_tag = 'DCStorageObject'

    def __init__(self, oma):
        super(XMLController, self).__init__()
        controller_element = oma.find('Controllers')
        if controller_element is None:
            return
            # TODO: See what multiple controllers looks like
        obj_element = controller_element.find(self.controller_obj_tag)
        self['name'] = obj_element.find('Name').text
        self['status'] = int(obj_element.find('ObjStatus').text)
        self['state'] = int(obj_element.find('ObjState').text)


class XMLPDisk(dict):
    pdisk_obj_tag = 'DCStorageObject'

    def __init__(self, oma):
        super(XMLPDisk, self).__init__()
        pdisk_element = oma.find('ArrayDisks')
        obj_elements = pdisk_element.findall(self.pdisk_obj_tag)
        for obj in obj_elements:
            device_id = int(obj.find('DeviceID').text)
            self[device_id] = dict()
            self[device_id]['status'] = int(obj.find('ObjStatus').text)
            self[device_id]['state'] = int(obj.find('ObjState').text)
            self[device_id]['vendor'] = obj.find('Vendor').text



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com, hussamd@gmail.com,
# prashantvm89@gmail.com, lee.scott@rackspace.com, Chris.Griffith@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import glob
import logging
import os

import lxml
import requests

from mercury_agent.capabilities import capability
from mercury_agent.inspector.inspect import global_device_info
from mercury.common.exceptions import HPFirmwareException
from mercury.common.helpers import cli
from mercury_agent.hardware.platform_detection import is_hp

log = logging.getLogger(__name__)

firmware_path = '/tmp/hp/firmware'
sum_log_path = '/var/log/sum/*.raw'


def vendor_is_hp():
    dmi_info = global_device_info.get('dmi')
    return is_hp(dmi_info)


def _extract(tarball_path, extract_path):
    if not os.path.exists(extract_path):
        os.makedirs(extract_path)
    log.info('Extracting {0}'.format(tarball_path))
    cmd = 'tar --strip-components=1 -xvf {0} -C {1}'.format(tarball_path,
                                                            extract_path)

    result = cli.run(cmd)
    if result.returncode:
        log.error('Could not extract archive: {0}'.format(result))
        return False

    return True


def parse_fw_report(xml_file):
    # noinspection PyUnresolvedReferences
    fw_tree = lxml.etree.parse(xml_file)
    # For checking needs_install XML tag is present or not
    component_list = parse_xml_for_components(fw_tree, 'needs_install')
    component_list += parse_xml_for_components(fw_tree, 'already_installed')
    return component_list


def parse_xml_for_components(xml_tree, xml_tag):
    def strip_date(version):
        return version.split('-')[0]

    component_list = []
    for i in xml_tree.findall(".//{0}/component".format(xml_tag)):
        installed = strip_date(i.find("installedversion").text)
        available = strip_date(i.find("version").text)
        component_name = i.find("name").text
        component_list.append(dict(name=component_name,
                                   installed=installed,
                                   available=available))
    return component_list


def find_available_updates():
    component_list = parse_fw_report(generate_fw_report())
    return firmware_updates_available(component_list)


def firmware_updates_available(component_list):
    update_flag = False
    for comp in component_list:
        log.debug('COMPONENT: {0}'.format(comp['name']))
        if comp['installed'].replace(' ', '') not in \
                comp['available'].replace(' ', ''):
            log.debug('CURRENT: {0}'.format(comp['installed']))
            log.debug('TARGET: {0}'.format(comp['available']))
            update_flag = True
        else:
            log.info('Firmware is up to date: {0}'.format(comp['available']))
    return update_flag


def _hpsum_cmd(cmd):
    # TODO: Figure out how to have cli.run() support timeouts
    # Agent probably already supports timeout for capabilities, verify!
    hpsum_exec_path = os.path.join(firmware_path, 'smartupdate')
    command = '{0} /s {1}'.format(hpsum_exec_path, cmd)
    return cli.run(command)


def generate_fw_report():
    result = _hpsum_cmd('/report')
    if result.returncode not in [0, 3]:
        raise HPFirmwareException('Unable to generate firmware report')

    xml_report = glob.glob(os.path.join(firmware_path, 'SUM*.xml'))
    if not xml_report:
        raise HPFirmwareException('Unable to find generated firmware report')

    return xml_report[0]


def get_return_code_status(ret):
    # Return codes for HP firmware updates.. supposedly for Linux
    ret_code = dict(
        HP_SUCCESS_NO_REBOOT=0,
        HP_SUCCESS_REBOOT=1,
        HP_SUCCESS_NOT_REQUIRED=3,
        HP_FAILURE_GENERAL=-1,
        HP_FAILURE_BAD_PARM=-2,
        HP_FAILURE_COMPONENT_FAILED=-3,
        HP_COMMAND_FAILURE=-4
    )

    for k in ret_code:
        if ret_code[k] == ret:
            ret = k

    return_code_status = dict(
        HP_SUCCESS_NO_REBOOT=dict(
            error_state=False,
            message='Firmware installation was successful and no reboot required',
            reboot_flag=False),
        HP_SUCCESS_REBOOT=dict(
            error_state=False,
            message='Firmware installation was successful and reboot is required',
            reboot_flag=True),
        HP_SUCCESS_NOT_REQUIRED=dict(
            error_state=False,
            message='Firmware installation successful and no update required',
            reboot_flag=False),
        HP_FAILURE_GENERAL=dict(
            error_state=True,
            message='General failure, check hpsum_log for errors',
            reboot_flag=False),
        HP_FAILURE_BAD_PARM=dict(
            error_state=True,
            message='Bad input parameter, check hpsum_log for errors',
            reboot_flag=False),
        HP_FAILURE_COMPONENT_FAILED=dict(
            error_state=True,
            message='Installation of component failed, check hpsum_logs for errors',
            reboot_flag=False),
        HP_COMMAND_FAILURE=dict(
            error_state=True,
            message='The CLI command execution failed',
            reboot_flag=False)
    )
    return return_code_status[ret]


def install_updates(force_install):
    log.info('Flashing firmware for this box')
    # Using /On_failed_dependency:Force parameter to avoid failed dependencies
    if force_install:
        log.info('Forcefully upgrading firmware. This may take sometime..')
        return _hpsum_cmd('/f /On_failed_dependency:Force')

    return _hpsum_cmd('/On_failed_dependency:Force')


def parse_raw_log():
    result = {
        'error': False,
        'exit_code': -99,
        'exit_message': 'Exit code failed to be parsed!',
        'reboot_required': False
    }

    raw_log_paths = glob.glob(sum_log_path)
    if not raw_log_paths:
        raise HPFirmwareException("Could not find smartupdate log file")

    with open(raw_log_paths[0]) as fp:
        raw_log_list = []
        for line in fp.readlines():
            split_line = line.strip().split('|')
            if 'Exit code' in split_line[-1]:
                exit_statement, exit_message = split_line[-1].split(',')
                result['exit_code'] = int(exit_statement.split(':')[1].strip())
                result['exit_message'] = exit_message.split(':')[1].strip()

                if result['exit_code'] == 1:
                    result['reboot_required'] = True

            raw_log_list.append('{} : {}'.format(split_line[0],
                                                 split_line[-1]))
    result['raw_log'] = raw_log_list
    return result


@capability('hp_apply_bios_settings',
            description='Apply bios settings found in given file',
            kwarg_names=['url'], serial=True,
            dependency_callback=vendor_is_hp, timeout=60)
def hp_apply_bios_settings(url=None):
    """
    Apply BIOS settings found at the given URL

    :param url: Full URL to the BIOS file
    """
    # TODO: Implement this
    # TODO: Alongside the conrepfile, some blobs have to be installed to bring
    # certain components up to a recommended version. Who determines which blobs
    # to execute? Are they provided or does this method
    # keep track of those based on the model
    ########
    # Consider writing a generic conrep interface, rather than relying on these
    # legacy interfaces. If the firmware needs to be updated prior to certain settings
    # being available, use the hp_update_firmware method first, rather than orchestrating
    # the upgrades here.
    # - Jared
    #########
    # DL380Gen9 = HPServer(name="ProLiant DL380 Gen9",
    #                      cfg='DL380Gen9_conrep_RAX.dat',
    #                      blobs=['BCM5719_2.13.5_RHEL6.scexe',
    #                             'HPSA_1.18_RHEL6.scexe',
    #                             'iLO4_2.02_RHEL6.scexe'])

    assert url
    raise NotImplementedError


@capability('hp_update_firmware',
            description='Installs updates from provided packages',
            kwarg_names=['url'], serial=True,
            dependency_callback=vendor_is_hp, timeout=3600)
def hp_update_firmware(url=None, dry_run=False):
    """
    Apply HP firmware updates

    :param url: Full URL to the package containing firmware files
    :param dry_run: If firmware updates should be applied when found
    """
    try:
        r = requests.get(url, stream=True, verify=False)
    except requests.RequestException as err:
        log.error('There was an error in retrieving the '
                  'firmware package: {}'.format(err))
        return

    download_path = '/tmp/hpfirmware.tar.gz'
    with open(download_path, 'wb') as f:
        for chunk in r.iter_content(1024 ** 2):
            if chunk:
                f.write(chunk)

    if not _extract(download_path, firmware_path):
        return

    updates_available = find_available_updates()
    if dry_run:
        if updates_available:
            log.info('Updates were available but were not installed')
        return

    # Return code could be a false positive, check log file output
    install_updates(updates_available)
    return parse_raw_log()



import threading


class AsyncInspector(object):
    def __init__(self, t):
        self.t = t
        self.kill = False

    def start(self, target, *args, **kwargs):
        self.t = threading.Thread(target=target, args=args, kwargs=kwargs)
        self.t.start()

    def inspect(self):
        raise NotImplementedError



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Module to unit test mercury_agent.inspector.hwlib.network_interfaces"""

import mock
import pytest

import netifaces

import mercury_agent.inspector.hwlib.network_interfaces as net_ifs
from tests.unit.base import MercuryAgentUnitTest


FAKE_INTERFACE_LIST = ['lo', 'eno1', 'br0', 'virbr0', 'virbr0-nic', 'vnet0']


def fake_ifaddresses_func(interface):
    """Fake netifaces.ifaddresses()."""
    ifaddresses = {
        'lo': {
            netifaces.AF_LINK: [{'addr': '93:b9:b7:be:3c:00'}],
            netifaces.AF_INET: [{'addr': '127.0.0.1'}],
            netifaces.AF_INET6: [{'addr': '::1'}]
        },
        'eno1': {
            netifaces.AF_LINK: [{'addr': 'd4:a8:28:a2:19:2f'}],
            netifaces.AF_INET: [{'addr': '192.168.1.115'}],
            netifaces.AF_INET6: [{'addr': 'dead:beef::1'}]
        },
        'virbr0': {
            netifaces.AF_LINK: [{'addr': 'a8:b5:61:1c:de:39'}],
            netifaces.AF_INET: [{'addr': '192.168.122.1'}],
            netifaces.AF_INET6: [{'addr': 'dead:beed::1'}]
        }
    }
    return ifaddresses.get(interface, {})


class MercuryMiscNetworkInterfacesUnitTests(MercuryAgentUnitTest):
    """Unit tests for mercury_agent.inspector.hwlib.network_interfaces"""
    def setUp(self):
        super(MercuryMiscNetworkInterfacesUnitTests, self).setUp()

    @mock.patch("netifaces.gateways")
    def test_get_default_interface(self, gateways_mock):
        """Tests for get_default_interface()"""
        gateways_mock.return_value = {}
        assert net_ifs.get_default_interface() == ''

        gateways_mock.return_value = {
            'default': {2: ('192.168.1.1', 'eth0')},
            2: [('192.168.1.1', 'eth0', True)]
        }
        assert net_ifs.get_default_interface() == 'eth0'

    @mock.patch("netifaces.ifaddresses")
    @mock.patch("netifaces.interfaces")
    def test_get_link_addresses(self, interfaces_mock, ifaddresses_mock):
        """Tests for get_link_addresses()"""
        interfaces_mock.return_value = FAKE_INTERFACE_LIST
        ifaddresses_mock.side_effect = fake_ifaddresses_func

        expected_output = [
            {'interface': 'lo', 'mac_address': '93:b9:b7:be:3c:00'},
            {'interface': 'eno1', 'mac_address': 'd4:a8:28:a2:19:2f'},
            {'interface': 'virbr0', 'mac_address': 'a8:b5:61:1c:de:39'},
        ]

        assert net_ifs.get_link_addresses(exclude_loopback=False) == \
            expected_output

    @mock.patch("netifaces.ifaddresses")
    @mock.patch("netifaces.interfaces")
    def test_get_ip4_network_info(self, interfaces_mock, ifaddresses_mock):
        """Tests for get_ipv4_network_info()."""
        interfaces_mock.return_value = FAKE_INTERFACE_LIST
        ifaddresses_mock.side_effect = fake_ifaddresses_func

        assert net_ifs.get_ipv4_network_info('not_an_interface') == []

        assert net_ifs.get_ipv4_network_info('eno1') == \
            [{'addr': '192.168.1.115'}]

    @mock.patch("netifaces.ifaddresses")
    @mock.patch("netifaces.interfaces")
    def test_get_ip6_network_info(self, interfaces_mock, ifaddresses_mock):
        """Tests for get_ipv6_network_info()."""
        interfaces_mock.return_value = FAKE_INTERFACE_LIST
        ifaddresses_mock.side_effect = fake_ifaddresses_func

        assert net_ifs.get_ipv6_network_info('not_an_interface') == []

        assert net_ifs.get_ipv6_network_info('virbr0') == \
            [{'addr': 'dead:beed::1'}]



import logging
import multiprocessing
import time

from mercury.common.exceptions import MercuryCritical

from mercury_agent.capabilities import capability
from mercury_agent.backend_client import get_backend_client
from mercury_agent.hardware.drivers.drivers import get_subsystem_drivers
from mercury_agent.hardware.erase import os_level_erase as os_erase
from mercury_agent.inspector.inspect import global_device_info
from mercury_agent.inspector.inspectors.os_storage import os_storage_inspector
from mercury_agent.inspector.inspectors.raid import raid_inspector

log = logging.getLogger(__name__)

HARDWARE_ERASE_SUPPORTED = [
    "PERC H740P Adapter"
]


def update_inventory():
    backend_client = get_backend_client()
    raid_info = raid_inspector(global_device_info)
    os_storage_info = os_storage_inspector()
    mercury_id = global_device_info['mercury_id']

    log.debug('Configuration changed, updating inventory and cache')
    update = {'raid': raid_info, 'os_storage': os_storage_info}
    global_device_info.update(update)
    backend_client.update(mercury_id, update)


def get_raid_drivers():
    """ Get raid driver targets for erase operations """
    hw_erase_drivers = []
    os_erase_drivers = []

    for driver in get_subsystem_drivers('raid'):
        for idx, _ in enumerate(driver.devices):
            if (driver.handler.get_adapter_info(idx)['name'].strip()
                    in HARDWARE_ERASE_SUPPORTED):
                hw_erase_drivers.append(driver)
            else:
                os_erase_drivers.append(driver)

    return hw_erase_drivers, os_erase_drivers


def driver_clear_all(drivers):
    for driver in drivers:
        for idx, _ in enumerate(driver.devices):
            log.info('Clearing configuration [%s adapter: %s]',
                     driver.name, idx)
            driver.handler.clear_configuration(idx)
    update_inventory()


def driver_start_erase(drivers, method):
    for driver in drivers:
        for idx, _ in enumerate(driver.devices):
            log.info('Starting Hardware Erase [%s adapter: %s]',
                     driver.name, idx)
            driver.handler.start_erase(idx, 'all', method)


def _assemble_jbod(driver, adapter):
    unassigned = driver.handler.get_unassigned(adapter)
    for idx, drive in enumerate(unassigned):
        log.info('Creating RAID 0 on %s disk %s', driver.name, idx)
        driver.handler.create_logical_drive(adapter, level='0', drives=[idx])


def driver_assemble_jbod(drivers):
    for driver in drivers:
        for idx, _ in enumerate(driver.devices):
            _assemble_jbod(driver, idx)
    update_inventory()


def _erase_wrapper(target, method, results_queue):
    # TODO: Figure out solution for remote logging and mp
    # noinspection PyBroadException
    try:
        results_queue.put(os_erase.erase(target, method))
    except Exception as e:
        results_queue.put({
            'error': True,
            'data': 'An error occurred while erasing device: {} [{}]'.format(
                    target,
                    e)
        })


def _background_erase(target, method):
    log.info('Backgrounding erase job for %s', target)
    results_queue = multiprocessing.Queue()
    p = multiprocessing.Process(target=_erase_wrapper, args=(
        target, method, results_queue))
    p.start()
    return {
        'process': p,
        'queue': results_queue,
        'target': target,
        'result': None,
        'completed': False
    }


def _check_hw_erase_completed(drivers):
    in_progress = 0
    for driver in drivers:
        for idx, _ in enumerate(driver.devices):
            res = driver.handler.erase_in_progress(idx)
            if res:
                # TODO: MRC-119, for now, we'll blast a log message
                # Only works for storcli
                for drive in driver.handler.get_erase_status(idx):
                    log.info('HW drive erase: %s [%s%%]', drive['Drive-ID'],
                             drive['Progress%'])
                in_progress += 1
    return in_progress == 0


def _check_os_erase_completed(os_erase_tasks):
    incomplete_tasks = [
        task for task in os_erase_tasks if not task['completed']]
    for incomplete in incomplete_tasks:
        process = incomplete['process']
        queue = incomplete['queue']
        if not process.is_alive():
            incomplete['completed'] = True
            incomplete['result'] = queue.get(block=False)
            log.info('Drive wipe operation has completed on %s',
                     incomplete['target'])
    return len(incomplete_tasks) == 0


def loop_until_completed(hw_erase_drivers, os_erase_tasks, interval):
    hw_erase_complete = False if hw_erase_drivers else True
    os_erase_complete = False if os_erase_tasks else True
    while True:
        if not hw_erase_complete:
            hw_erase_complete = _check_hw_erase_completed(hw_erase_drivers)
        if not os_erase_complete:
            os_erase_complete = _check_os_erase_completed(os_erase_tasks)
        if hw_erase_complete and os_erase_complete:
            log.info('Background jobs completed!')
            break
        time.sleep(interval)


def _check_background_errors(os_erase_tasks):
    has_errors = False
    for task in os_erase_tasks:
        if task['result'].get('error'):
            has_errors = True
            log.error('Error erasing drive: %s', task)

    if has_errors:
        raise MercuryCritical('An error has occurred, the system has not been '
                              'erased')


@capability('erase',
            'Erase all data from local storage',
            timeout=0,
            serial=True)
def erase(method='fast'):
    """

    :param method:
    :return:
    """

    check_interval = 30

    log.info('Checking device for local storage')
    hw_method = 'simple' if method == 'fast' else 'standard'
    os_method = 'zero' if method == 'fast' else 'unsupported'
    hw_erase_drivers, os_erase_drivers = get_raid_drivers()

    # clear all RAID configurations
    log.info('Clearing all adapter RAID configurations')
    driver_clear_all(hw_erase_drivers + os_erase_drivers)

    # Start Hardware Erase
    if hw_erase_drivers:
        log.info('Initiating hardware erase on supported controllers')
        driver_start_erase(hw_erase_drivers, hw_method)

    # Assemble JBOD for wiping
    if os_erase_drivers:
        log.info('Assembling managed drives into JBOD')
        driver_assemble_jbod(os_erase_drivers)

    storage_devices = global_device_info['os_storage']

    # Wipe SSDs using TRIM
    ssd_results = []
    ssd_devices = [drive['udev']['DEVNAME'] for drive in storage_devices
                   if drive['media_type'] == 'ssd']
    for drive in ssd_devices:
        log.info('Preparing to TRIM erase SSDs')
        ssd_results.append(os_erase.erase(drive, 'ssd_trim'))

    # TODO: Add periodic status updates
    background_erase_tasks = []
    log.info('Initiating background wipe of remaining devices')
    for drive in [drive['udev']['DEVNAME'] for drive in storage_devices
                  if drive['media_type'] == 'disk']:
        background_erase_tasks.append(_background_erase(drive, os_method))

    loop_until_completed(hw_erase_drivers, background_erase_tasks,
                         check_interval)

    # clear all RAID configurations
    log.info('Clearing all adapter RAID configurations')
    driver_clear_all(hw_erase_drivers + os_erase_drivers)

    _check_background_errors(background_erase_tasks)
    log.info('System erase has completed')

    return {'report': {
        'ssd': ssd_results,
        'rotational': [t['result'] for t in background_erase_tasks]
    }}



import logging

from mercury.common.exceptions import MercuryGeneralException
from mercury.common.clients.router_req_client import RouterReqClient

LOG = logging.getLogger('')


class MercuryLogHandler(logging.Handler):
    def __init__(self, service_url, mercury_id=None):
        super(MercuryLogHandler, self).__init__()

        self.service_url = service_url
        self.__mercury_id = mercury_id

        self.client = RouterReqClient(self.service_url, linger=0,
                                      response_timeout=2)
        self.service_name = 'Logging Service'

    def emit(self, record):
        if not self.__mercury_id:
            return

        data = record.__dict__
        data.update({'mercury_id': self.__mercury_id})

        err_msg = 'There is a problem with the remote logging service!'
        # noinspection PyBroadException
        try:
            response = self.client.transceiver(data)
        except Exception:
            logging.error(err_msg)
        else:
            if response.get('error'):
                logging.error(err_msg)
        finally:
            self.client.close()

    def set_mercury_id(self, mercury_id):
        self.__mercury_id = mercury_id



# Copyright 2018 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury_agent.capabilities import capability
from mercury_agent.backend_client import get_backend_client
from mercury_agent.hardware.drivers.drivers import driver_class_cache
from mercury_agent.inspector.inspect import global_device_info
from mercury_agent.inspector.inspectors.raid import raid_inspector


log = logging.getLogger(__name__)


def get_megaraid_driver():
    return driver_class_cache.get('megaraid_sas')


def has_megaraid_driver():
    return bool(get_megaraid_driver())


def update_inventory():
    backend_client = get_backend_client()
    raid_info = raid_inspector(global_device_info)
    mercury_id = global_device_info['mercury_id']

    log.debug('RAID configuration changed, updating inventory')
    backend_client.update(mercury_id, {'raid': raid_info})


def update_on_change(f):
    def wrapped_f(*args, **kwargs):
        result = f(*args, **kwargs)
        update_inventory()
        return result
    wrapped_f.__name__ = f.__name__
    wrapped_f.__doc__ = f.__doc__
    return wrapped_f


@capability('megaraid_add',
            description='Create and array on a megaraid_sas based controller',
            kwarg_names=['controller', 'array_type', 'drives'],
            serial=True,
            dependency_callback=has_megaraid_driver,
            timeout=60,
            task_id_kwargs=True
            )
@update_on_change
def megaraid_add(controller,
                 array_type,
                 drives,
                 size=None,
                 pdperarray=None,
                 pdcache=None,
                 dimmerswitch=None,
                 io_mode='direct',
                 write_policy='wb',
                 read_policy='ra',
                 cachevd=False,
                 stripe_size=None,
                 spares=None,
                 cached_bad_bbu=False,
                 after_vd=None):
    """ Add virtual drive

    :param controller: Controller ID
    :param array_type:  r[0|1|5|6|10|50|60]
    :param drives: Drives specified with as EncID:drive,...
    :param size: Size of a drive in MB or None for maximum
    :param pdperarray: Specifies the number of physical drives per array. The
default value is automatically chosen.(0 to 16)
    :param pdcache: Enables or disables PD cache. (on|off|default)
    :param dimmerswitch: Specifies the power-saving policy. Sets to default
automatically. default: Logical device uses controller default power-saving
policy. automatic (auto): Logical device power savings managed by firmware.
none: No power-saving policy.
maximum (max): Logical device uses maximum power savings.
MaximumWithoutCaching (maxnocache): Logical device does not cache write to
maximize power savings.
    :param io_mode: cached|direct
    :param write_policy:wb|rt
    :param read_policy:ra|rt
    :param cachevd: enables or disables cachecade device support
    :param stripe_size: stripe size ( the amount of data writen before moving to
the next disk )
    :param spares: Numer drives allocated as hot spares
    :param cached_bad_bbu: Enable write caches even when the bbu is missing or
discharged
    :param after_vd: Specify an existing VD to add this new vd behind
    :return: AttributeString of command output
    """

    driver = get_megaraid_driver()

    log.info('Adding Array: /c{} RAID{} drives: {} size: {}'.format(
        controller,
        array_type,
        drives,
        size
    ))

    return driver.storcli.add(
        controller,
        array_type,
        drives,
        size=size,
        pdperarray=pdperarray,
        pdcache=pdcache,
        dimmerswitch=dimmerswitch,
        io_mode=io_mode,
        write_policy=write_policy,
        read_policy=read_policy,
        cachevd=cachevd,
        stripe_size=stripe_size,
        spares=spares,
        cached_bad_bbu=cached_bad_bbu,
        after_vd=after_vd,
    )


@capability('megaraid_delete',
            description='Delete megaraid based virtual drive',
            kwarg_names=['controller'],
            serial=True,
            dependency_callback=has_megaraid_driver,
            timeout=60,
            task_id_kwargs=True)
@update_on_change
def megaraid_delete(controller, virtual_drive='all'):
    """

    :param controller: Controller id or all
    :param virtual_drive: Virtual Drive id or all
    :return:
    """

    driver = get_megaraid_driver()
    log.info('Deleting virtual drive: {} on controller: {}'.format(
        controller, virtual_drive
    ))

    return driver.storcli.delete(controller, virtual_drive)



class AsyncInspector(object):
    def __init__(self, device_info, agent_configuration):
        self.device_info = device_info
        self.configuration = agent_configuration

    def inspect(self):
        """Cannot block"""
        raise NotImplemented

    def cleanup(self):
        """Called on agent exit"""
        raise NotImplemented


# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
From the pyudev docs:

    Once added, a filter cannot be removed anymore. Create a new object instead.

pyudev can be kind of silly, but I certainly don't feel like wrapping my own.
"""
import logging
import os
import pyudev

log = logging.getLogger(__name__)


class UDevHelper(object):
    def __init__(self):
        self.context = pyudev.Context()

    def get_monitor(self):
        """
        Because filters cannot be removed
        :return: fresh pyudev.Monitor
        """
        return pyudev.Monitor.from_netlink(self.context)

    def get_partitions(self):
        return self.context.list_devices(subsystem='block', DEVTYPE='partition')

    def get_disks(self):
        return self.context.list_devices(subsystem='block', DEVTYPE='disk')

    def find_partitions(self, device):
        """
        matches partitions belonging to a device.
        """

        devices = self.get_partitions()
        return devices.match_parent(pyudev.Device.from_device_file(self.context, device))

    def get_device_by_name(self, devname):
        try:
            udisk = pyudev.Device.from_device_file(self.context, devname)
        except OSError:
            return None
        return udisk

    def discover_valid_storage_devices(self, fc_enabled=True, loop_enabled=False):
        """
        Kind of ugly, but gets the job done. It strips devices we don't
        care about, such as cd roms, device mapper block devices, loop, and fibre channel.

        """

        disks = self.get_disks()
        pruned = list()

        for disk in disks:
            if not fc_enabled and 'fc' in disk.get('ID_BUS', ''):
                continue

            if not loop_enabled and disk.get('MAJOR') == '7':
                continue

            if disk.get('ID_TYPE') == 'cd':
                continue

            if disk.get('MAJOR') == '254':  # Device Mapper (LVM)
                continue

            if os.path.split(disk.get('DEVPATH', ''))[-1].startswith('ram'):
                continue

            pruned.append(disk)

        return pruned

    def yield_mapped_devices(self):
        disks = self.get_disks()
        for disk in disks:
            if disk.get('MAJOR') == '254':  # Device Mapper (LVM)
                yield disk

    @staticmethod
    def monitor_partition_by_devname(monitor, partition_id, action=None):
        monitor.filter_by('block', device_type="partition")
        for _, device in monitor:
            log.debug('Seen: %s' % list(device.items()))
            if action and device.get('ACTION') != action:
                log.debug('Action, %s, does not match %s' % (action, device.get('ACTION')))
                continue
            if device.get('UDISKS_PARTITION_NUMBER') == str(partition_id):
                return str(device['DEVNAME'])

    def get_network_devices(self):
        """ Returns a list of all network(ethernet/type 1] devices found on the system. """

        result = []

        for candidate in self.context.list_devices(subsystem='net'):
            try:
                if candidate.attributes.asint('type') == 1:
                    result.append(candidate)
            except KeyError:
                pass

        # let's go ahead and return it sorted..
        result.sort(key=lambda dev: dev.sys_name)
        return result

    @staticmethod
    def monitor_for_volume(monitor, lv_name):
        monitor.filter_by('block')
        for action, device in monitor:
            if device.get('DM_LV_NAME') == lv_name:
                return device



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os


def build_index(l, key):
    our_dict = dict()
    for d in l:
        if key not in d:
            continue
        idx = d[key]
        if idx in our_dict:
            our_dict[idx].append(d)
        else:
            our_dict[idx] = [d]
    return our_dict


def get_cpufreq_info(cpu):
    sys_cpu_path = '/sys/devices/system/cpu/cpu%s/cpufreq' % str(cpu)
    if not os.path.exists(sys_cpu_path):
        return dict()

    def read(path):
        with open(path) as fp:
            return int(fp.read().strip())

    freq = dict()
    freq['min'] = read(os.path.join(sys_cpu_path, 'scaling_min_freq'))
    freq['max'] = read(os.path.join(sys_cpu_path, 'scaling_max_freq'))
    freq['cur'] = read(os.path.join(sys_cpu_path, 'scaling_cur_freq'))

    return freq


class CPUInfo(object):
    def __init__(self):
        if not os.path.exists('/proc/cpuinfo'):
            raise OSError('/proc/cpuinfo is missing. Bro, do you even linux?')

        with open('/proc/cpuinfo') as fp:
            self.raw_cpuinfo = fp.read()

        cores = self.raw_cpuinfo.split('\n\n')

        self.core_dicts = list()
        for core in cores:
            if not core:
                continue
            core_dict = dict()
            for attribute in core.splitlines():
                if not attribute:
                    continue
                k, v = attribute.split(':')
                fixed_key = k.strip().replace(' ', '_').lower()
                stripped_value = v.strip()

                if fixed_key in ['processor', 'physical_id', 'core_id', 'cpu_cores']:
                    stripped_value = int(stripped_value)

                core_dict[fixed_key] = stripped_value

            self.core_dicts.append(core_dict)

        self.core_dicts.sort(key=lambda d: d['processor'])

    @property
    def physical_index(self):
        return build_index(self.core_dicts, 'physical_id')

    @property
    def logical_processor_index(self):
        return build_index(self.core_dicts, 'processor')

    @property
    def processor_ids(self):
        return [core_dict['processor'] for core_dict in self.core_dicts]

    @property
    def physical_processor_count(self):
        return len(self.physical_index)

    @property
    def logical_core_count(self):
        #  AKA, threads (HTT)
        return len(self.core_dicts)

    @property
    def total_physical_core_count(self):
        #  We assume that multi socket systems will be using the same proc
        return self.cores_per_processor * self.physical_processor_count

    def get_cores(self, physical_id):
        return self.physical_index.get(physical_id)

    @property
    def cores_per_processor(self):
        return self.one_core.get('cpu_cores')

    @property
    def core_zero_index(self):
        physical_index = self.physical_index
        for physical_id in physical_index:
            physical_index[physical_id] = physical_index[physical_id][0]
        return physical_index

    @staticmethod
    def get_speed_info(core_dict):
        speed_info = dict()
        processor_id = int(core_dict['processor'])
        speed_info['model_name'] = core_dict['model_name']
        cpufreq = get_cpufreq_info(processor_id)
        cpufreq_enabled = bool(cpufreq) or False

        speed_info['bogomips'] = float(core_dict['bogomips'])

        speed_info['cpufreq_enabled'] = cpufreq_enabled

        if cpufreq_enabled:
            speed_info['current'] = float(cpufreq['cur'])
            speed_info['min'] = float(cpufreq['min'])
            speed_info['max'] = float(cpufreq['max'])
        else:
            cpu_speed = core_dict['cpu_mhz']
            speed_info['current'] = float(cpu_speed)
            speed_info['min'] = float(cpu_speed)
            speed_info['max'] = float(cpu_speed)

        return speed_info

    def get_physical_speed_info(self):
        speed_info = list()
        zero_index = self.core_zero_index
        for physical_processor in zero_index:
            core_dict = zero_index[physical_processor]
            speed_info.append(self.get_speed_info(core_dict))
        return speed_info

    @property
    def one_core(self):
        return self.core_dicts and self.core_dicts[0] or dict()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

from mercury.common.configuration import MercuryConfiguration


log = logging.getLogger(__name__)

# Helpers
AGENT_CONFIG_FILE = 'mercury-agent.yaml'


__configuration = {}


def parse_options():
    configuration = MercuryConfiguration(
        'mercury-agent',
        AGENT_CONFIG_FILE,
        description='The Mercury Agent'
    )

    configuration.add_option('agent.bind_address',
                             help_string='The interface and port for socket '
                                         'binding',
                             default='tcp://127.0.0.1:9003'),

    configuration.add_option('agent.pong_bind_address',
                             help_string='Interface and port for the pong'
                                         'service',
                             default='tcp://127.0.0.1:9004')

    configuration.add_option('agent.dhcp_ip_source',
                             help_string='The method of determining dhcp_ip',
                             default='simple',
                             one_of=['simple', 'udhcpd', 'routing_table'])

    configuration.add_option('agent.remote.backend_url',
                             help_string='The ZeroMQ URL of the backend '
                                         'service',
                             required=True)

    configuration.add_option('agent.remote.log_service_url',
                             help_string='Optional logging service zURL')

    configuration.add_option('agent.hardware.raid.storcli_path',
                             cli_argument='--storcli_path',
                             env_variable='STORCLI_PATH',
                             default='storcli64')

    configuration.add_option('agent.hardware.raid.hpssacli_path',
                             cli_argument='--hpssacli_path',
                             env_variable='HPSSACLI_PATH',
                             default='hpssacli')

    configuration.add_option('agent.local_ip',
                             cli_argument='--local-ip',
                             env_variable='MERCURY_AGENT_ADDRESS',
                             help_string='The address which will be advertised'
                                         'to the backend for communication with'
                                         'this agent')
    configuration.add_option('agent.pong_log_level',
                             cli_argument='--pong-log-level',
                             default='ERROR',
                             help_string='The pong process log level')

    configuration.add_option('agent.hardware.obm.racadm_path',
                             cli_argument='--racadm-path',
                             env_variable='RACADM_PATH',
                             help_string='The location of the racadm binary',
                             default='racadm')

    configuration.add_option('agent.hardware.megacli_bin',
                             help_string='megacli binary',
                             default='/usr/local/sbin/megacli')

    configuration.add_option('agent.hardware.mce_threshold',
                             help_string='The maximum number of corrected machine '
                                         'check exceptions which can occur before '
                                         'an error is raised',
                             default=20,
                             special_type=int)

    configuration.add_option('agent.hardware.oem.hp.hpasmcli_path',
                             help_string='The path of the hpasmcli binary',
                             env_variable='HPASMCLI_PATH',
                             default='hpasmcli')

    return configuration.scan_options()


def get_configuration():
    global __configuration
    if not __configuration:
        __configuration = parse_options()
    return __configuration



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Module to unit test mercury_agent.inspector.hwlib.cpuinfo"""

import mock
import pytest

import mercury_agent.inspector.hwlib.cpuinfo as cpuinfo
from tests.unit.base import MercuryAgentUnitTest


EXAMPLE_PROC_CPUINFO_OUTPUT = """processor	: 0
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1268.115
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 0
cpu cores	: 6
apicid		: 0
initial apicid	: 0
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6612.06
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:


processor	: 1
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1199.835
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 1
cpu cores	: 6
apicid		: 2
initial apicid	: 2
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6615.66
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 2
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.036
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 2
cpu cores	: 6
apicid		: 4
initial apicid	: 4
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.18
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 3
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.036
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 3
cpu cores	: 6
apicid		: 6
initial apicid	: 6
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.25
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 4
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.640
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 4
cpu cores	: 6
apicid		: 8
initial apicid	: 8
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.27
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 5
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.439
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 5
cpu cores	: 6
apicid		: 10
initial apicid	: 10
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.28
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 6
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1300.543
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 0
cpu cores	: 6
apicid		: 1
initial apicid	: 1
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6618.23
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 7
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.439
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 1
cpu cores	: 6
apicid		: 3
initial apicid	: 3
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.75
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 8
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.439
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 2
cpu cores	: 6
apicid		: 5
initial apicid	: 5
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.36
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 9
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.036
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 3
cpu cores	: 6
apicid		: 7
initial apicid	: 7
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.22
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor	: 10
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1199.835
cache size	: 15360 KB
physical id	: 0
siblings	: 12
core id		: 4
cpu cores	: 6
apicid		: 9
initial apicid	: 9
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs		:
bogomips	: 6616.40
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

processor       : 11
vendor_id       : GenuineIntel
cpu family      : 6
model           : 63
model name      : Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping        : 2
microcode       : 0x39
cpu MHz         : 1201.245
cache size      : 15360 KB
physical id     : 0
siblings        : 12
core id         : 5
cpu cores       : 6
apicid          : 11
initial apicid  : 11
fpu             : yes
fpu_exception   : yes
cpuid level     : 15
wp              : yes
flags           : fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts
bugs            :
bogomips        : 6616.29
clflush size    : 64
cache_alignment : 64
address sizes   : 46 bits physical, 48 bits virtual
power management:

"""

CPUINFO_TEMPLATE_SINGLE_CPU_ENTRY = """processor	: %(processor_num)s
vendor_id	: GenuineIntel
cpu family	: 6
model		: 63
model name	: Intel(R) Core(TM) i7-5820K CPU @ 3.30GHz
stepping	: 2
microcode	: 0x39
cpu MHz		: 1200.640
cache size	: 15360 KB
physical id	: %(physical_id)s
siblings	: %(siblings)s
core id		: %(core_id)s
cpu cores	: %(cpu_cores)s
apicid		: %(apic_id)s
initial apicid	: %(apic_id)s
fpu		: yes
fpu_exception	: yes
cpuid level	: 15
wp		: yes
flags		: %(flags)s
bugs		:
bogomips	: 6616.29
clflush size	: 64
cache_alignment	: 64
address sizes	: 46 bits physical, 48 bits virtual
power management:

"""

DEFAULT_CPU_FLAGS = "fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm epb tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid cqm xsaveopt cqm_llc cqm_occup_llc dtherm ida arat pln pts"


def get_fake_cpuinfo_output(num_logical_processors=2,
                            hyper_threaded=False,
                            flags=DEFAULT_CPU_FLAGS):
    """Generates valid-looking cpuinfo output for testing."""
    fake_output = ''
    cpu_cores = (num_logical_processors / 2) if hyper_threaded else \
        num_logical_processors
    for proc_num in range(0, num_logical_processors):
        fake_output += CPUINFO_TEMPLATE_SINGLE_CPU_ENTRY % {
            'processor_num': proc_num,
            'physical_id': 0,
            'siblings': num_logical_processors,
            'core_id': (proc_num % 2) if hyper_threaded else proc_num,
            'cpu_cores': cpu_cores,
            'apic_id': proc_num,
            'flags': flags
        }
    return fake_output


class MercuryHwlibCpuinfoUnitTests(MercuryAgentUnitTest):
    """Unit tests for mercury_agent.inspector.hwlib.cpuinfo"""
    @mock.patch("mercury_agent.inspector.hwlib.cpuinfo.os.path.exists")
    @mock.patch("mercury_agent.inspector.hwlib.cpuinfo.open")
    def setUp(self, open_mock, os_path_exists_mock):
        """Setup a CPUInfo object to use for further tests."""
        super(MercuryHwlibCpuinfoUnitTests, self).setUp()
        os_path_exists_mock.return_value = True
        open_mock.return_value.__enter__.return_value.read.return_value = \
            EXAMPLE_PROC_CPUINFO_OUTPUT
        self.cpuinfo_obj = cpuinfo.CPUInfo()

    @mock.patch("mercury_agent.inspector.hwlib.cpuinfo.os.path.exists")
    def test_proc_cpuinfo_absent(self, os_path_exists_mock):
        """Test what happens if /proc/cpuinfo doesn't exist.

        Particularly when constrcuting a CPUInfo object.
        """
        os_path_exists_mock.return_value = False
        with pytest.raises(OSError):
            cpuinfo.CPUInfo()

    def test_cpu_info_example_output(self):
        """Look at various properties of cpuinfo object to check sanity.

        Tests against EXAMPLE_PROC_CPUINFO_OUTPUT, which is cpuinfo output
        for a 5820k (6 physical cores, 12 logical ones).
        """
        cores_zero = self.cpuinfo_obj.get_cores(0)
        assert isinstance(cores_zero, list)
        assert len(cores_zero) == 12
        for core in cores_zero:
            assert isinstance(core, dict)

        assert self.cpuinfo_obj.processor_ids == list(range(0, 12))
        assert self.cpuinfo_obj.logical_core_count == 12
        assert self.cpuinfo_obj.total_physical_core_count == 6
        assert self.cpuinfo_obj.cores_per_processor == 6

        lpi = self.cpuinfo_obj.logical_processor_index
        assert isinstance(lpi, dict)
        assert len(lpi.keys()) == 12
        assert list(lpi.keys()) == list(range(0, 12))

        core_zi = self.cpuinfo_obj.core_zero_index
        assert isinstance(core_zi, dict)
        assert len(core_zi.keys()) == 1

    def test_core_dict_missing_phys_id(self):
        """Test behavior when a core_dict is missing data."""
        del self.cpuinfo_obj.core_dicts[0]['physical_id']
        result = self.cpuinfo_obj.physical_index
        assert isinstance(result, dict)
        assert len(result.keys()) == 1

    @mock.patch("mercury_agent.inspector.hwlib.cpuinfo.os.path.exists")
    @mock.patch("mercury_agent.inspector.hwlib.cpuinfo.open")
    def test_get_cpufreq_info(self, open_mock, os_path_exists_mock):
        """Test get_cpufreq_info()."""
        # Test normal, expected operation.
        os_path_exists_mock.return_value = True
        open_mock.return_value.__enter__.return_value.read.side_effect = [
            "1200000", "4000000", "2000000"]

        result = cpuinfo.get_cpufreq_info("0")

        cpufreq_address = '/sys/devices/system/cpu/cpu0/cpufreq'
        assert os_path_exists_mock.called
        os_path_exists_mock.assert_called_with(cpufreq_address)
        open_mock.assert_has_calls([
            mock.call(cpufreq_address + '/scaling_min_freq'),
            mock.call(cpufreq_address + '/scaling_max_freq'),
            mock.call(cpufreq_address + '/scaling_cur_freq')], any_order=True)

        assert isinstance(result, dict)
        assert len(result.keys()) == 3
        assert result['min'] == 1200000
        assert result['max'] == 4000000
        assert result['cur'] == 2000000

        # See what happens if the sys cpu path doesn't exist.
        os_path_exists_mock.return_value = False
        result = cpuinfo.get_cpufreq_info("0")
        assert result == {}

    @mock.patch("mercury_agent.inspector.hwlib.cpuinfo.get_cpufreq_info")
    def test_get_speed_info(self, cpufreq_info_mock):
        """Test CPUInfo.get_physical_speed_info() operation."""
        cpufreq_info_mock.side_effect = [
            {'min': 1200000, 'max': 4000000, 'cur': 2000000}] * 12
        speed_info = self.cpuinfo_obj.get_physical_speed_info()
        assert isinstance(speed_info, list)
        assert len(speed_info) == 1
        assert speed_info[0]['cpufreq_enabled'] == True
        assert speed_info[0]['bogomips'] == 6612.06
        assert speed_info[0]['current'] == 2000000.00
        assert speed_info[0]['min'] == 1200000.00
        assert speed_info[0]['max'] == 4000000.00

        # See what happens when cpufreq is not available.
        cpufreq_info_mock.side_effect = [False]
        speed_info = self.cpuinfo_obj.get_physical_speed_info()
        assert isinstance(speed_info, list)
        assert len(speed_info) == 1
        assert speed_info[0]['cpufreq_enabled'] == False
        assert speed_info[0]['bogomips'] == 6612.06
        assert speed_info[0]['current'] == 1268.115
        assert speed_info[0]['min'] == 1268.115
        assert speed_info[0]['max'] == 1268.115



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com, hussamd@gmail.com,
# prashantvm89@gmail.com, lee.scott@rackspace.com, Chris.Griffith@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os

from mercury_agent.capabilities import capability
from mercury_agent.hardware.platform_detection import is_dell
from mercury_agent.inspector.inspect import global_device_info
from mercury.common.exceptions import MercuryFirmwareException
from mercury.common.helpers.util import extract_tar_archive, download_file, cli, xml_to_dict


log = logging.getLogger(__name__)
firmware_path = '/tmp/dell/firmware'


def vendor_is_dell():
    dmi_info = global_device_info.get('dmi')
    return is_dell(dmi_info)


def get_return_code_status(ret):
        # Return codes for DELL firmware updates.. supposedly for Linux
        ret_code = dict(
            SUCCESS_NO_REBOOT=0,
            FAILURE=1,
            SUCCESS_REBOOT_REQUIRED=2,
            DEP_SOFT_ERROR=3,
            DEP_HARD_ERROR=4,
            QUAL_HARD_ERROR=5,
            REBOOTING_SYSTEM=6,
            RPM_VERIFY_FAILED=9,
            UPDATE_SUCCESSFUL_SOFT_DEPEDENCY_NOT_MET=13,
            REBOOT_REQUIRED_SOFT_DEPENDENCY_NOT_MET=14
        )

        for k in ret_code:
            if ret_code[k] == ret:
                ret = k

        return_code_status = dict(
            SUCCESS_NO_REBOOT=dict(
                error_state=False,
                message='Update was successful and no reboot required',
                reboot_flag=False),
            FAILURE=dict(
                error_state=True,
                message='Update was unsuccessful (Failure)',
                reboot_flag=False),
            SUCCESS_REBOOT_REQUIRED=dict(
                error_state=False,
                message='Update was successful and reboot is required',
                reboot_flag=True),
            DEP_SOFT_ERROR=dict(
                error_state=True,
                message='Soft dependency error',
                reboot_flag=False),
            DEP_HARD_ERROR=dict(
                error_state=True,
                message='Hard dependency errors',
                reboot_flag=False),
            QUAL_HARD_ERROR=dict(
                error_state=True,
                message='Qualification error',
                reboot_flag=False),
            REBOOTING_SYSTEM=dict(
                error_state=False,
                message='Rebooting system',
                reboot_flag=True),
            RPM_VERIFY_FAILED=dict(
                error_state=True,
                message='RPM verification has failed',
                reboot_flag=False),
            UPDATE_SUCCESSFUL_SOFT_DEPEDENCY_NOT_MET=dict(
                error_state=False,
                message='Update is successful. Soft Dependencies are not met.',
                reboot_flag=False),
            REBOOT_REQUIRED_SOFT_DEPENDENCY_NOT_MET=dict(
                error_state=False,
                message='Update is successful. Reboot is required. Soft Dependencies are not met.',
                reboot_flag=True)
        )
        return return_code_status[ret]


def run_cmd_and_parse_xml(cmd, xml_element):
    result = cli.run(cmd).stdout
    return xml_to_dict(result, xml_element)


def _split_and_get_fw_vers(xml_element_dict, xml_ele):
    fw_vers = []
    for xml_element in xml_element_dict:
        split_value = str(xml_element[xml_ele]).split()
        if not split_value:
            continue
        # DRAC special case
        if 'FWVersion' in xml_ele:
            fw_vers.append(split_value[0])
        else:
            fw_vers.append(split_value[1 if len(split_value) > 1 else 0])
    return set(fw_vers)


def _get_runner(args, xml_element, xml_tag):
    command = 'omreport {} -fmt xml'.format(args)
    xml_element_dict = run_cmd_and_parse_xml(command, xml_element)
    return _split_and_get_fw_vers(xml_element_dict, xml_tag)


def get_nics_fw():
    return _get_runner('chassis nics', './/DevNicObj', 'FirmwareVersion')


def get_storage_controller_fw():
    return _get_runner('storage controller', './/DCStorageObject', 'FirmwareVer')


def get_bios_fw():
    return _get_runner('chassis bios', './/SystemBIOS', 'Version')


def get_drac_fw():
    return _get_runner('chassis firmware', './/Firmware', 'FWVersion')


def generate_fw_report():
    return {
        'BIOS': get_bios_fw(),
        'DRAC': get_drac_fw(),
        'Storage Controller': get_storage_controller_fw(),
        'Network Adapter': get_nics_fw()
    }


def parse_fw_report(existing_firmware):
    component_list = []
    for component_name, installed_versions in existing_firmware.items():
        for ver in installed_versions:
            component_list.append({
                'name': component_name,
                'version': ver,
                'is_installed': any(
                    (ver in installer)
                    for installer in os.listdir(firmware_path)
                )
            })
    return component_list


def firmware_updates_available(component_list):
    update_flag = False
    for comp in component_list:
        log.info(('COMPONENT: {0}'.format(comp['name'])))
        if comp['is_installed']:
            log.info('Firmware is up to date {0}'.format(comp['version']))
        else:
            log.info('CURRENT: {0}'.format(comp['version']))
            log.info('TARGET: Higher version is available')
            update_flag = True
    return update_flag


def find_available_updates():
    # omreport command works without sudo
    component_list = parse_fw_report(generate_fw_report())
    return firmware_updates_available(component_list)


def _check_all():
        update_list = list()
        check_args = ' -q -c'

        for installer in os.listdir(firmware_path):
            check_cmd = os.path.join(firmware_path, installer)
            check_cmd = check_cmd + check_args
            result = cli.run(check_cmd)
            if result.returncode == 5:
                log.info('{0} is not applicable'.format(installer))
            elif result.returncode == 3:
                log.info('{0} is already updated'.format(installer))
            else:
                log.info('{0} update must run'.format(installer))
                update_list.append(installer)
        return update_list


def install_updates():
    reboot_required = False
    log.debug('Flashing firmware for this box')
    update_list = _check_all()
    if not update_list:
        log.debug('All firmware is up to date')
        return reboot_required

    errors = []
    for installer in update_list:
        update_cmd = os.path.join(firmware_path, installer)
        update_cmd = update_cmd + ' -q'
        log.info('Running update: {0}'.format(installer))
        result = cli.run(update_cmd)
        status = get_return_code_status(int(result.returncode))
        if status.get('error_state'):
            err_msg = '{0}: {1}'.format((status.get('message')), result)
            log.error(err_msg)
            errors.append(err_msg)
        else:
            reboot_required = True
    if errors:
        raise MercuryFirmwareException("Errors while flashing Dell Firmware: "
                                       "{}".format(", ".join(errors)))
    return reboot_required


@capability('dell_apply_bios_settings',
            description='Apply bios settings found in given file',
            kwarg_names=['url'], serial=True,
            dependency_callback=vendor_is_dell, timeout=600)
def dell_apply_bios_settings(url=None):
    """
    Apply BIOS settings found at the given URL

    :param url: Full URL to the BIOS file
    """
    # TODO: Implement this
    assert url
    raise NotImplementedError


@capability('dell_update_firmware',
            description='Installs updates from provided packages',
            kwarg_names=['url', 'dry_run'], serial=True,
            dependency_callback=vendor_is_dell, timeout=3600)
def dell_update_firmware(url, dry_run=False):
    """
     Apply HP firmware updates
     :param url: Full URL to the package containing firmware files
     :param dry_run: If firmware updates should be applied when found
    """
    download_path = '/tmp/dell_firmware.tar.gz'

    download_file(url, download_path)

    result = extract_tar_archive(download_path, firmware_path)
    if result.returncode:
        raise MercuryFirmwareException('Could not extract firmware file: {}'.format(result))

    updates_available = find_available_updates()
    if dry_run:
        if updates_available:
            log.info('Updates were available but were not installed')
        return

    # Return code could be a false positive, check log file output
    reboot_required = install_updates()
    # TODO: Implement a reboot capability?
    return dict(result=dict(return_code=1, success=True, reboot_required=reboot_required))



import unittest

import box
import mock

from mercury_agent.inspector.inspectors import health

_config = {
    'agent':
        {
            'hardware':
                {
                    'mce_threshold': 10,
                    'oem':
                        {
                            'hp':
                                {
                                    'hpasmcli_path': 'noop'
                                }
                        }
                }
        }
}


class SysHealthInspector(unittest.TestCase):
    @mock.patch('mercury_agent.inspector.inspectors.health.mcelog')
    @mock.patch('mercury_agent.inspector.inspectors.health.hpasmcli')
    @mock.patch('mercury_agent.inspector.inspectors.health.get_configuration')
    def test_with_hp_error(self, mock_get_config, mock_hpasm, mock_mce):
        device_info = {'dmi': {'sys_vendor': 'HP'}}

        class FakeHP:
            def show_server(self):
                return {'processors': [{'status': 'Nope'}]}

            def show_dimm(self):
                return []

            def show_powersupply(self):
                return []

        mock_mce.count_logged_events = mock.Mock(return_value=0)
        mock_get_config.return_value = box.Box(_config)
        mock_hpasm.HPASMCLI = mock.Mock(return_value=FakeHP())
        mock_hp_obj = mock_hpasm.HPASMCLI()
        print(mock_hp_obj.show_server())

        print(health.system_health_inspector(device_info))



import unittest

import mock
from mercury.common.helpers.cli import CLIResult

from mercury_agent.hardware.oem.hp import hpasmcli


class TestHPASMCLI(unittest.TestCase):
    show_server_data = '\n\nSystem        : ProLiant DL380 Gen9\nSerial No.    : TC51NR9952\n' \
               'ROM version   : v2.60 (05/21/2018) P89\nUEFI Support  : Yes\niLo ' \
               'present   : Yes\nEmbedded NICs : 8\n\tNIC1 MAC: 38:63:bb:3f:4b:f4\n\t' \
               'NIC2 MAC: 38:63:bb:3f:4b:f5\n\tNIC3 MAC: 38:63:bb:3f:4b:f6\n\tNIC4 MAC: ' \
               '38:63:bb:3f:4b:f7\n\tNIC5 MAC: 8c:dc:d4:ad:d6:d0\n\tNIC6 MAC: ' \
               '8c:dc:d4:ad:d6:d1\n\tNIC7 MAC: 68:05:ca:39:89:a0\n\tNIC8 MAC: 68:05:ca:39:89:a1' \
               '\n\nProcessor: 0\n\tName         : Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz' \
               '\n\tStepping     : 2\n\tSpeed        : 2400 MHz\n\tBus          : 100 MHz\n\tCore' \
               '         : 8\n\tThread       : 16\n\tSocket       : 1\n\tLevel1 Cache : ' \
               '512 KBytes\n\tLevel2 Cache : 2048 KBytes\n\tLevel3 Cache : 20480 KBytes\n\t' \
               'Status       : Ok\n\nProcessor: 1\n\tName         : ' \
               'Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz\n\tStepping     : 2\n\t' \
               'Speed        : 2400 MHz\n\tBus          : 100 MHz\n\tCore         : 8\n\t' \
               'Thread       : 16\n\tSocket       : 2\n\tLevel1 Cache : 512 KBytes\n\t' \
               'Level2 Cache : 2048 KBytes\n\tLevel3 Cache : 20480 KBytes\n\tStatus       : Ok' \
               '\n\nProcessor total  : 2\n\nMemory installed : 131072 MBytes\nECC ' \
               'supported    : Yes\n\n\n'
    show_dimm_data = """
DIMM Configuration
------------------
Processor #:                     1
Module #:                     1
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     1
Module #:                     4
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     1
Module #:                     9
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     1
Module #:                     12
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     2
Module #:                     1
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     2
Module #:                     4
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     2
Module #:                     9
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok

Processor #:                     2
Module #:                     12
Present:                      Yes
Form Factor:                  9h
Memory Type:                  DDR4(1ah)
Size:                         16384 MB
Speed:                        2133 MHz
Supports Lock Step:           No
Configured for Lock Step:     No
Status:                       Ok




"""
    show_power_supply_data = '\nPower supply #1\n\tPresent  : Yes\n\tRedundant: Yes' \
                             '\n\tCondition: Ok\n\tHotplug  : Supported\n\tPower    : 110 ' \
                             'Watts\nPower supply #2\n\tPresent  : Yes\n\tRedundant: Yes' \
                             '\n\tCondition: Ok\n\tHotplug  : Supported\n\tPower    : 100 Watts\n\n'

    def setUp(self):
        self.mock_cli_patch = mock.patch('mercury_agent.hardware.oem.hp.hpasmcli.cli')
        self.mock_cli = self.mock_cli_patch.start()
        self.mock_cli.find_in_path = mock.Mock(return_value='/bin/bash')

    def tearDown(self):
        self.mock_cli_patch.stop()

    def test_show_server(self):
        self.mock_cli.run = mock.Mock(return_value=CLIResult(
            self.show_server_data, '', 0))

        hpasm = hpasmcli.HPASMCLI()
        details = hpasm.show_server()

        self.assertEqual(len(details['processors']), 2)
        self.assertEqual(details['system'], 'ProLiant DL380 Gen9')

    def test_show_dimmm(self):
        self.mock_cli.run = mock.Mock(return_value=CLIResult(
            self.show_dimm_data, '', 0))
        hpasm = hpasmcli.HPASMCLI()
        details = hpasm.show_dimm()

        self.assertEqual(details[-1]['processor_#'], 2)

    def test_show_power_supply(self):
        self.mock_cli.run = mock.Mock(return_value=CLIResult(
            self.show_power_supply_data, '', 0))

        hpasm = hpasmcli.HPASMCLI()
        power_supplies = hpasm.show_powersupply()
        for ps in power_supplies:
            self.assertEqual(ps['condition'], 'Ok')



from size import PercentString, Size


class RAIDAbstractionException(Exception):
    pass


class RAIDActions(object):
    """
    Bare bones RAID abstraction layer. Simple support for querying and modifying a RAID controller
    through a standard interface. Many things are not supported, but may be added at a later date
    If controller specific advanced features are needed, see the provider APIs and RPC capabilities

    """

    # TODO add RAID constraints class object for use in raid_minimums and raid_calculator

    def transform_adapter_info(self, adapter_index):
        raise NotImplementedError

    def _add_indexes(self, configuration):
        """
        This method is called to add drive indexes to transformed configuration. This is necessary
        because drives are now only present within an adapter configuration, but we need still
        need abstractly map drives physical location. This is achieved by calling instance.sort_drives
        on a list of aggregate drives, enumerating them, and storing the enumeration as drive.index

        :param configuration: configuration reference of instance.transform_adapter_info output
        :return: None
        """

        drives = [pd for array in configuration['arrays'] for pd in array['physical_drives']]

        drives += (configuration['spares'] + configuration['unassigned'])
        self.sort_drives(drives)

        for idx, drive in enumerate(drives):
            drive.update({'index': idx})

    def _add_totals(self, adapter_info):
        """
        Adds drive and size totals to adapter_info
        :param adapter_info:
        :return:
        """
        all_drives = self.get_all_drives_from_adapter(adapter_info)
        adapter_info['total_size'] = sum(
            [drive['size'] for drive in all_drives])
        adapter_info['total_drives'] = len(all_drives)

    def get_adapter_info(self, adapter_index):
        """
        Returns adapter details in a standard format. The driver is responsible for finding
        adapters to perform actions on. The abstraction uses indexes rather than vendor references.

        For instance, hpsa uses pci slot ids to specify adapters, onboard adapters will be 0,
        expansion cards will be 3 and 4, should they occupy pci-e slots 3 and 4 on the riser card.
        In order to 'flatten' the interface, these cards will be accessed by enumeration:

        adapter0: slot0
        adapter1: slot3
        adapter2: slot4

        In the rare occurrence that there are multiple manufacturers, cards will enumerated based
        on there subsystem ids (PCI Ids for example). To implement this, the driver probe process
        will need to change so that drivers and handlers are registered in PCI order and not import
        order. Talk to Jared R. about implementation details as he is soldiering on.

        :param adapter_index: The index of the adapter to populate configuration information
        :return: A dictionary representing the adapter

        """
        adapter_info = self.transform_adapter_info(adapter_index)
        self._add_indexes(adapter_info['configuration'])
        self._add_totals(adapter_info)
        return adapter_info

    def create(self, adapter_info, level, drives=None, size=None, array=None):
        """
        This method should only be called from the base classes create_logical_volume method.
        This method will be implemented by all Actions classes.

        :param adapter_info: Transformed adapter data
        :param level: A supported RAID level
        :type level: str
        :param drives: A zero based, expanded list of participating drive indexes.
            Should be None if array is specified
        :param size: A python-size object or None (maximum)
        :param array: The index of the existing array (with sufficient free space) being targeted
            Should be None if drives are specified
        :return:
        """
        raise NotImplementedError

    def create_logical_drive_on_existing_array(self, adapter, adapter_info, level, array, converted_size,
                                               percent_string):
        """

        :param adapter:
        :param adapter_info:
        :param level:
        :param array:
        :param converted_size:
        :param percent_string:
        :return:
        """

        # Check to see that the array exists
        try:
            array = adapter_info['configuration']['arrays'][array]
        except (KeyError, IndexError):
            raise RAIDAbstractionException('The referenced array {}:{} does not exist'.format(adapter, array))

        # Check to see if the array has enough free space (or any free space, if size is None)
        free_space = Size(array['free_space'])
        if free_space < Size('1MiB'):
            raise RAIDAbstractionException('The array {}:{} has not free space'.format(adapter, array))

        if converted_size:
            if free_space < converted_size:
                raise RAIDAbstractionException('Requested size {} exceeds available size {}'.format(
                    converted_size, free_space))
        elif percent_string:
            if percent_string.free:
                converted_size = Size(percent_string.value * free_space.bytes)
            else:
                raise RAIDAbstractionException('only %FREE is supported for RAID abstraction')

        return self.create(adapter_info, level, drives=None, size=converted_size, array=array)

    def create_logical_drive_on_new_array(self, adapter, adapter_info, level, drives, converted_size, percent_string):
        """

        :param adapter:
        :param adapter_info:
        :param level:
        :param drives:
        :param converted_size:
        :param percent_string:
        :return:
        """
        # Creating new array, so we need to check that there is enough space (if specified)
        # and that the RAID level is valid
        target_drives = self.get_drives_from_selection(adapter, drives)

        # this will raise an exception if it's invalid
        self.raid_minimums(level, len(target_drives))

        # Now we have to get the smallest drive (though they all should be the same)

        if converted_size or percent_string:
            smallest_drive_size = sorted([x['size'] for x in target_drives])[0]

            available_size = self.raid_calculator(level, len(target_drives), smallest_drive_size)
            if converted_size:
                if available_size < converted_size.bytes:
                    raise RAIDAbstractionException('Requested size of {} exceeds available size of {}'.format(
                        converted_size, Size(available_size)))
            else:
                converted_size = Size(percent_string.value * available_size)

        return self.create(adapter_info, level, drives=target_drives, size=converted_size, array=None)

    def create_logical_drive(self, adapter, level, drives=None, size=None, array=None):
        """
        :param adapter:
        :type adapter: int
        :param level: 0, 1, 5, 6, 10, 1+0, 50, 60
        :type level: str
        :param drives: drives should be referenced as 0 based comma separated indexes, ranges, or
        a combination of both. When using a range, both the lower and upper bounds are inclusive.
        Another option is to use all or unassigned. `all` requires that all drives on the adapter
        are not part of the array. `unassigned` will select all drives not currently members of an
        array or participating as spares
        :param size: Size can be specified in bytes (int), a string using SI or IEC standards, a
        percent of total space, or a percent of free space available. If no size is listed, all
        available space will be used
        :param array: An index of an existing array we are updating
        """

        # TODO: Add span support
        # TODO: Add generic cache policy support

        if not (array is not None or drives is not None):
            raise RAIDAbstractionException('Either drive targets or an array must be specified')

        adapter_info = self.get_adapter_info(adapter)

        percent_string = None
        converted_size = None

        if size:
            if isinstance(size, str) and '%' in size:
                percent_string = PercentString(size)
            else:
                converted_size = Size(size)

        if drives is not None:
            return self.create_logical_drive_on_new_array(adapter,
                                                          adapter_info,
                                                          level,
                                                          drives,
                                                          converted_size,
                                                          percent_string)
        else:
            return self.create_logical_drive_on_existing_array(adapter,
                                                               adapter_info,
                                                               level,
                                                               array,
                                                               converted_size,
                                                               percent_string)

    def delete_logical_drive(self, adapter, array, logical_drive):
        raise NotImplementedError

    def clear_configuration(self, adapter):
        raise NotImplementedError

    def add_spares(self, adapter, drives, arrays=None):
        raise NotImplementedError

    def get_all_drives_from_adapter(self, adapter):
        """
        Get drives while preserving array membership
        :param adapter: target adapter
        :return: drives
        :return type: list
        """
        drives = []

        arrays = adapter['configuration']['arrays']
        for array_idx in range(len(arrays)):
            this_array = arrays[array_idx]
            for pd in this_array['physical_drives']:
                pd.update({'member_of': array_idx})
                drives.append(pd)

        drives += adapter['configuration']['spares'] \
                  + adapter['configuration']['unassigned']

        self.sort_drives(drives)

        return drives

    def get_all_drives(self, adapter_index):
        """
        Get drives while preserving array membership
        :param adapter_index: target adapter index
        :return: drives
        :return type: list
        """

        return self.get_all_drives_from_adapter(
            self.get_adapter_info(adapter_index))

    def get_unassigned(self, adapter_index):
        drives = self.get_all_drives(adapter_index)

        pruned = []
        for drive in drives:
            if 'member_of' in drive:
                continue
            if 'target' in drive:
                continue
            pruned.append(drive)

        return pruned

    @staticmethod
    def sort_drives(drives):
        drives.sort()

    @staticmethod
    def get_selection_from_pattern(pattern):
        """
        The pattern is a mechanism to target one or more drives attached to a controller.
        See the create_logical_volume documentation for the full format
        :param pattern:
        :type pattern: str
        :return: a list of indexes
        """

        invalid_msg = 'Pattern {} is invalid'.format(pattern)
        items = pattern.split(',')
        selection = []

        for item in items:
            if item.strip().isdigit():
                selection.append(int(item))
                continue

            if '-' not in item:
                raise RAIDAbstractionException(invalid_msg)

            our_range = item.split('-')

            if len(our_range) != 2:
                raise RAIDAbstractionException(invalid_msg)

            try:
                front, back = [int(x) for x in our_range]
            except ValueError:
                raise RAIDAbstractionException('Range is nonsense: {}'.format(our_range))

            if not back > front:
                raise RAIDAbstractionException(invalid_msg + ' [Range is negative]')

            selection += list(range(front, back + 1))

        return selection

    def parse_selection(self, adapter_index, drive_selector):
        # single drive expressed as an index
        if isinstance(drive_selector, int):
            return [drive_selector]

        # a short circuited list
        if isinstance(drive_selector, list):
            return drive_selector

        if drive_selector.lower() == 'all':
            return [drive['index'] for drive in self.get_all_drives(adapter_index)]

        if drive_selector.lower() == 'unassigned':
            return [drive['index'] for drive in self.get_unassigned(adapter_index)]

        return sorted(self.get_selection_from_pattern(drive_selector))

    def fetch_selection(self, adapter_index, selection):
        # Build an index of unassigned drives for O(1) matching
        unassigned_drives = {
            d['index']: d for d in self.get_unassigned(adapter_index)
            }

        selected_drives = []

        for drive in selection:
            ud = unassigned_drives.get(drive)
            if not ud:
                raise RAIDAbstractionException('Drive {} is not available'.format(drive))

            if not ud['status'] == 'OK':
                raise RAIDAbstractionException(
                    'Attempting to initialize a failed drive : {} {}'.format(
                        drive, ud['status']))
            selected_drives.append(ud)
        return selected_drives

    def get_drives_from_selection(self, adapter_index, drive_selector):
        selection = self.parse_selection(adapter_index, drive_selector)
        return self.fetch_selection(adapter_index, selection)

    # These methods can be overridden to support more raid levels
    # TODO: Clean these up...
    @staticmethod
    def raid_calculator(level, number_of_drives, size_of_drive, span_depth=2):
        if level == '0':
            return size_of_drive * number_of_drives
        if level == '1':
            return size_of_drive
        if level == '5':
            return size_of_drive * (number_of_drives - 1)
        if level == '6':
            return size_of_drive * (number_of_drives - 2)
        if level == '10' or level == '1+0':
            return size_of_drive * (number_of_drives / span_depth)
        if level == '50':
            return size_of_drive * (number_of_drives - 2)
        if level == '60':
            return size_of_drive * (number_of_drives - 4)

        raise RAIDAbstractionException('Unsupported RAID level')

    @staticmethod
    def raid_minimums(level, number_of_drives):
        if level == '1' and number_of_drives < 2:
            raise RAIDAbstractionException('RAID 1 requires at least 2 drives')
        elif level == '5' and number_of_drives < 3:
            raise RAIDAbstractionException('RAID 5 requires at least 3 drives')
        elif level == '6' and number_of_drives < 4:
            raise RAIDAbstractionException('RAID 6 requires at least 4 drives')
        elif (level == '10' or level == '1+0') and number_of_drives < 4:
            raise RAIDAbstractionException('RAID 10 requires at least 4 drives')
        elif level == '50' and number_of_drives < 6:
            raise RAIDAbstractionException('RAID 50 requires at least 6 drives')
        elif level == '60' and number_of_drives < 8:
            raise RAIDAbstractionException('RAID 60 requires at least 8 drives')

        if level in ['10', '1+0', '50', '60'] and number_of_drives % 2:
            raise RAIDAbstractionException('RAID10/50/60 require an even number of drives')

    def erase(self, adapter, drives, method='fast'):
        """

        :param adapter:
        :param drives:
        :param method:
        :return:
        """
        raise NotImplementedError

    def get_erase_status(self, adapter):
        """

        :param adapter:
        :return:
        """
        raise NotImplementedError

    def erase_in_progress(self, adapter):
        raise NotImplementedError



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Temporary baseline driver for LSI/PERC MegaRAID SAS adapters
# Next iteration will harness stor/percCLI which has JSON output capabilities

import logging

from size import Size

from mercury_agent.configuration import get_configuration

from mercury_agent.hardware import platform_detection
from mercury_agent.hardware.drivers import driver, PCIDriverBase
from mercury_agent.hardware.raid.abstraction.api import RAIDActions, \
    RAIDAbstractionException
from mercury_agent.hardware.raid.interfaces.megaraid.storcli import Storcli

log = logging.getLogger(__name__)


class MegaRAIDActions(RAIDActions):
    # Drives have many more statuses than we care about
    # This information will be preserved in extra
    drive_status_map = {
        'DHS': 'OK',  # Dedicated Hotspare
        'UGood': 'OK',
        'GHS': 'OK',  # Global Hotspare
        'UBad': 'Failed',
        'Onln': 'OK',
        'Offln': 'Failed'
    }

    logical_drive_map = {
        'Rec': 'RECOVERING',
        'Pdgd': 'DEGRADED',
        'dgrd': 'DEGRADED',
        'Optl': 'OK'
    }

    hotspare_map = {
        'DHS': 'dedicated',
        'GHS': 'global'
    }

    def __init__(self):
        """ MegaRAID support for RAID abstraction.
        This class is using the mercury native storcli interface. The interface is very thin.
        As such, vendor_info may need a little more cleanup in comparison to SmartArray
        """
        super(MegaRAIDActions, self).__init__()
        self.storcli = Storcli(binary_path=get_configuration().get(
            'agent', {}).get(
            'hardware', {}).get(
            'raid', {}).get(
            'storcli_path') or 'storcli')

    @staticmethod
    def get_vendor_info(adapter):
        """

        :param adapter:
        :return:
        """
        return {
            'general': adapter['Basics'],
            'version_info': adapter['Version'],
            'bus': adapter['Bus'],
            'status': adapter['Status'],
            'supported_adapter_ops': adapter['Supported Adapter Operations'],
            'supported_pd_ops': adapter['Supported PD Operations'],
            'supported_vd_ops': adapter['Supported VD Operations'],
            'bbu_info': adapter.get('BBU_Info')
        }

    @staticmethod
    def dash_is_none(value):
        """

        :param value:
        :return:
        """
        if value != '-':
            return value

    @staticmethod
    def get_controller_id(adapter_info):
        return adapter_info['vendor_info']['general']['Controller']

    def populate_logical_drives(self, disk_group, dg_info):
        """

        :param disk_group:
        :param dg_info:
        :return:
        """
        lds = []
        for vd in dg_info['VD LIST']:
            dg_id, vd_id = (int(x) for x in vd['DG/VD'].split('/'))
            if dg_id == disk_group:
                lds.append({
                    'level': int(vd['TYPE'].strip('RAID')),
                    'size': Size(vd['Size'], force_iec_values=True).bytes,
                    'status': self.logical_drive_map.get((vd['State']),
                                                         'UNKNOWN'),
                    'extra': {
                        'disk_group': dg_id,
                        'access': vd['Access'],
                        'consistency': vd['Consist'],
                        'cache': vd['Cache'],
                        'name': vd['Name'],
                        'vendor_state': vd['State']
                    }
                })

        return lds

    def format_physical_drive(self, pd):
        """

        :param pd:
        :return:
        """
        # DG JSON type is inconsistent
        disk_group = self.dash_is_none(pd['DG'])
        if isinstance(disk_group, str):
            disk_group = int(disk_group)

        return {
            'size': Size(pd['Size'], force_iec_values=True).bytes,
            'status': self.drive_status_map.get(pd['State'], 'UNKNOWN'),
            'type': pd['Med'],
            'extra': {
                'address': pd['EID:Slt'],
                'drive_id': pd['DID'],
                'model': pd['Model'],
                'disk_group': disk_group,
                'interface': pd['Intf'],
                'self_encrypted_drive':
                    pd['SED'] != 'N',
                'spun': pd['Sp'],
                'sector_size': Size(pd['SeSz'].lower()).bytes,
                'vendor_state': pd['State'],
                'spare_type': self.hotspare_map.get(pd['State'], None)
            }
        }

    def populate_physical_drives(self, disk_group, dg_info):
        """

        :param disk_group:
        :param dg_info:
        :return:
        """
        pds = []
        for pd in dg_info['DG Drive LIST']:
            if pd['DG'] == disk_group:
                pds.append(self.format_physical_drive(pd))

        return pds

    @staticmethod
    def populate_free_space(disk_group, dg_info):
        """

        :param disk_group:
        :param dg_info:
        :return:
        """
        free_space_details = dg_info.get('FREE SPACE DETAILS')

        if free_space_details:
            for fs in free_space_details:
                if fs['DG'] == disk_group:
                    return Size(fs['Size'], force_iec_values=True).bytes

        return 0

    def get_configuration(self, dg_info):
        """ Builds an array of storage array structures for use

        :param dg_info:
        :return:
        """
        arrays = []

        topology = dg_info['TOPOLOGY']

        # The JSON we are working with is horrid
        # Rather than nesting, disk group structures are signified by
        # containing a '-' in the Arr (Array) field. Further, I do not
        # believe there is a a mechanism to create more than one array
        # per disk group. In this abstraction, a disk group IS an array.
        dgs = [dg for dg in topology if dg['Arr'] == '-'
               and dg['Type'] != 'DRIVE']

        for dg in dgs:
            arrays.append({
                'logical_drives': self.populate_logical_drives(dg['DG'],
                                                               dg_info),
                'physical_drives': self.populate_physical_drives(dg['DG'],
                                                                 dg_info),
                'free_space': self.populate_free_space(dg['DG'],
                                                       dg_info),
                'extra': {
                    'disk_group': dg['DG'],
                    'background_task': dg['BT'] == 'yes',
                    'dimmer_switch': self.dash_is_none(dg['DS3'])
                }
            })

        return arrays

    @staticmethod
    def get_array_index_by_dg(arrays, dg):
        """

        :param arrays: Transformed arrays object
        :param dg: Disk group target
        :return:
        """
        cnt = 0
        for array in arrays:
            if array['extra']['disk_group'] == dg:
                return cnt
            cnt += 1
        return -1

    def get_spares(self, arrays, dg_info):
        """

        :param arrays:
        :param dg_info:
        :return:
        """
        drive_list = dg_info.get('UN-CONFIGURED DRIVE LIST')

        if not drive_list:
            return []

        spares = []

        for drive in drive_list:
            if drive['State'] in self.hotspare_map:
                spare = self.format_physical_drive(drive)

                disk_group = spare['extra']['disk_group']
                if disk_group is not None:
                    # DHS
                    spare['target'] = self.get_array_index_by_dg(arrays,
                                                                 disk_group)
                else:
                    spare['target'] = None

                spares.append(spare)

        return spares

    def get_unconfigured_drives(self, dg_info):
        """

        :param dg_info:
        :return:
        """
        drive_list = dg_info.get('UN-CONFIGURED DRIVE LIST')

        if not drive_list:
            return []

        drives = []

        for drive in drive_list:
            if drive['State'] not in self.hotspare_map:
                drives.append(self.format_physical_drive(drive))

        return drives

    def transform_configuration(self, controller):
        """

        :param controller:
        :return:
        """
        configuration = {
            'arrays': [],
            'spares': [],
            'unassigned': []
        }

        # The second controller specific read is necessary because
        # /call show all does not contain some necessary information.
        # Such as "UN-CONFIGURED DRIVE LIST"
        dg_info = self.storcli.get_disk_group(controller)[0]

        if dg_info.get('TOPOLOGY'):
            # A configuration exists on the adapter
            configuration['arrays'] = self.get_configuration(dg_info)
            configuration['spares'] = self.get_spares(configuration['arrays'],
                                                      dg_info)

        configuration['unassigned'] = self.get_unconfigured_drives(dg_info)

        return configuration

    @staticmethod
    def sort_drives(drives):
        drives.sort(key=lambda d: d['extra']['drive_id'])

    def transform_adapter_info(self, adapter_index):
        """
        Transforms Storcli.controllers[adapter_index] into standard form
        :param adapter_index:
        :return: Adapter details in standard from
        """
        try:
            adapter = self.storcli.controllers[adapter_index]
        except IndexError:
            raise RAIDAbstractionException('Controller does not exist')

        adapter_details = {
            'name': adapter['Basics']['Model'],
            'provider': 'megaraid',
            'vendor_info': self.get_vendor_info(adapter),
            'configuration': self.transform_configuration(adapter_index)
        }

        return adapter_details

    def create(self, adapter_info, level, drives=None, size=None, array=None):
        """

        :param adapter_info:
        :param level:
        :param drives:
        :param size:
        :param array:
        :return:
        """
        size_mb = size and round(size.megabytes) or None

        level = 'r{}'.format(level)

        target_drives = []
        for drive in drives:
            if drive['extra']['vendor_state'] != 'UGood':
                raise RAIDAbstractionException(
                    'Drive index: {}, address:{} is in a failed state, cannot '
                    'create array'.format(
                        drive['index'], drive['extra']['address']))

            target_drives.append(drive['extra']['address'])

        return self.storcli.add(
            controller=self.get_controller_id(adapter_info),
            array_type=level,
            drives=','.join(target_drives),
            size=size_mb
        )

    def delete_logical_drive(self, adapter, array, logical_drive):
        """

        :param adapter:
        :param array:
        :param logical_drive:
        :return:
        """
        adapter_info = self.get_adapter_info(adapter)
        controller_id = self.get_controller_id(adapter_info)

        arrays = adapter_info['configuration']['arrays']

        try:
            # ensure that the index exists
            assert len(arrays[array]['logical_drives']) >= logical_drive + 1
        except (IndexError, AssertionError):
            raise RAIDAbstractionException(
                'Logical Drive does not exist at {}:{}:{}'.format(
                    adapter, array, logical_drive)
            )
        return self.storcli.delete(controller=controller_id,
                                   virtual_drive=logical_drive)

    def clear_configuration(self, adapter):
        """

        :param adapter:
        :return:
        """
        return self.storcli.delete(controller=self.get_controller_id(
            self.get_adapter_info(adapter)), virtual_drive='all')

    def add_spares(self, adapter, drives, arrays=None):
        """

        :param adapter:
        :param drives:
        :param arrays:
        :return:
        """
        adapter_info = self.get_adapter_info(adapter)
        controller_id = self.get_controller_id(adapter_info)

        target_drives = self.get_drives_from_selection(adapter, drives)

        results = []
        dgs = []

        if arrays:
            for ary_index in arrays:
                try:
                    array = adapter_info['configuration']['arrays'][ary_index]
                except IndexError:
                    raise RAIDAbstractionException(
                        'Array {} does not exist'.format(ary_index))
                dgs.append(array['extra']['disk_group'])

        for drive in target_drives:
            enclosure, slot = (
                int(_) for _ in drive['extra']['address'].split(':'))
            results.append(
                self.storcli.add_hotspare(controller=controller_id,
                                          enclosure=enclosure,
                                          slot=slot,
                                          disk_groups=dgs)
            )

        return results

    def erase(self, adapter, drives, method='fast'):
        log.info(
            'Starting hardware erase on adapter: %s, drives: %s, method: %s',
            adapter, drives, method)

        adapter_info = self.get_adapter_info(adapter)
        controller_id = self.get_controller_id(adapter_info)

        if drives == 'all':
            self.storcli.start_erase(controller_id, 'all', 'all')
        else:
            enclosure_drive_map = {}
            target_drives = self.get_drives_from_selection(0, drives)
            for _drive in target_drives:
                enclosure, slot = (int(_) for _ in _drive['extra']['address'].split(':'))
                if enclosure not in enclosure_drive_map:
                    enclosure_drive_map[enclosure] = [slot]
                else:
                    enclosure_drive_map[enclosure].append(slot)

            for enclosure in enclosure_drive_map.keys():
                self.storcli.start_erase(
                    controller_id, enclosure, ','.join(enclosure_drive_map[enclosure]))

    def get_erase_status(self, adapter):
        log.info('Getting hardware erase status on adapter: %s, drives: %s',
                 adapter)
        adapter_info = self.get_adapter_info(adapter)
        controller_id = self.get_controller_id(adapter_info)

        return self.storcli.show_erase(controller_id, 'all', 'all')

    def erase_in_progress(self, adapter):
        log.info('Checking erase status')
        for drive in self.get_erase_status(adapter):
            if drive['Status'] == 'In Progress':
                return True
        return False


@driver()
class MegaRaidSASDriver(PCIDriverBase):
    name = 'megaraid_sas'  # named after the kernel module
    driver_type = 'raid'
    _handler = MegaRAIDActions
    wants = 'pci'

    @classmethod
    def probe(cls, pci_data):
        raid_pci_devices = platform_detection.get_raid_controllers(pci_data)
        owns = list()

        for device in raid_pci_devices:
            if cls.check(device):
                owns.append(device['slot'])
        return owns

    @classmethod
    def check(cls, pci_device):
        return pci_device['driver'] == cls.name

    def inspect(self):
        adapters = []

        for idx in range(len(self.devices)):
            adapters.append(self.handler.get_adapter_info(idx))

        return adapters



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import time

from mercury_agent.configuration import get_configuration
from mercury.common.exceptions import MercuryCritical, MercuryConfigurationError
from mercury_agent.inspector.inspectors.interfaces import get_interface_by_name
from mercury_agent.inspector.inspectors.routes import find_default_route

log = logging.getLogger(__name__)


def get_dhcp_ip(device_info, method='simple'):
    """
    Get the ip address provided by the DHCP server. Initial operation will use this IP to register with the rpc
     service. Plans are in the works for providing multiple avenues for selecting/generating(ipv6) addresses. Some
     options are:
        Command line argument
        Kernel command line arguments
        Link layer addressing, sans DHCP (48bit, IPV6 only : RFC2373)
    :param device_info: mercury.agent.inspector device_info
    :param method: Possible methods are:
                    simple - Find the interface used for the default route, select this interfaces primary (first)
                    address
                    udhcpc - A mercury_ghost and mercury.agent.inspector construct. The ghost (ephemeral) os that
                    carries the agent and inspector codes contains a script that is called during udhcpc operation.
                    This script will configured the interface as normal. In addition the script will write a file,
                    mercury_dhcp_info.sh, containing DHCP options recieved from the server (including the dhcp_ip).
    :return: ip address
    """
    override_ip = get_configuration()['agent']['local_ip']
    if override_ip:
        return override_ip

    if method == 'simple':
        routes = device_info['routes']
        interfaces = device_info['interfaces']
        default_route = find_default_route(routes)
        if not default_route:
            raise MercuryCritical('The is no default route defined, cannot use simple method')
        if 'src' in default_route:
            return default_route['src']

        # If the route does not contain a src definition, we take the first address on the 'default' dev
        default_interface = get_interface_by_name(interfaces, default_route['dev']) or dict()
        address_info = default_interface.get('address_info')
        if not address_info:
            raise MercuryCritical('Failed to determine dhcp address')

        log.debug('address_info: %s' % address_info)
        return address_info[0]['addr']


def _serialize_capabilities(capabilities):
    _d = {}
    for c in capabilities:
        temp_d = capabilities[c].copy()
        temp_d['entry'] = temp_d['entry'].__name__
        callback = temp_d['dependency_callback']
        if callback and not callback():
            log.debug('{} : Dependency callback returned false, not exposing capability'.format(c))
            continue
        del temp_d['dependency_callback']
        _d[c] = temp_d
        log.debug('{} : Capability confirmed and exportable'.format(c))
    return _d


def _get_port_from_config(key, default):
    """ Hacky way to avoid having another configuration option

    :param key: service_bind_address or pong_bind_address
    :param default: The default port 9003 or 9004 respectively
    :return: The port
    """

    zurl = get_configuration().agent.get(key)
    if not zurl:
        return default

    # obviously, things could go wrong here
    try:
        return int(zurl.split(':')[2].strip())
    except (ValueError, KeyError, IndexError):
        raise MercuryConfigurationError(f'{key} seems to be invalid in config'
                                        f'zURL format is required:'
                                        f'proto://address:port')


def register(rpc_backend_client, device_info, local_ip, local_ip6, capabilities):
    # There is still some confusion regarding how best to determine what ip
    # to publish. Current wisdom suggest that we find the default gateway,
    # determine the interface used for the default route, and take whatever
    # address is configured. This 'should' be the DHCP address right?
    # Relying solely on dhclient (in the initrd) might be an option and would
    # allow for parsing dhclient.leases.
    # Currently, the preboot environment uses udhcpc, so modifying the script
    # udhcpc calls on renew/bound to publish a cookie with vars set by the dhcp
    # server is an option as well

    # Either process should be configurable. ie,
    #  default_ip_source: simple # use default route
    #                     udhcpc # script in preconfig leaves cookie, parse the cookie
    #                     dhclient # parse /var/lib/dhclient.leases

    # And then there is ipv6, where we'd need to DHCP to get a route-able address
    # subnet, and then heuristically generate an ip address in the subnet, forgoing
    # the explicit need for publishing the address

    # UPDATE: Determine the route taken to find rpc_backend and use that interface

    agent_info = {
        'rpc_address': local_ip,
        'rpc_address6': local_ip6,
        'rpc_port': _get_port_from_config('service_bind_address', 9003),
        'ping_port': _get_port_from_config('pong_bind_address', 9004),
        'localtime': time.time(),
        'capabilities': _serialize_capabilities(capabilities)
    }

    return rpc_backend_client.register(device_info, agent_info)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import time

from mercury_agent.capabilities import capability
from mercury_agent.configuration import get_configuration
from mercury.common.clients.rpc.backend import BackEndClient
from mercury.common.exceptions import fancy_traceback_short, parse_exception

from press.configuration.util import configuration_from_file
from press.exceptions import PressCriticalException
from press.log import setup_logging
from press.plugin_init import init_plugins
from press.press import PressOrchestrator
from press.press_cli import parse_args
from press.hooks.hooks import clear_hooks


log = logging.getLogger(__name__)


def cleanup_thread():
    # Clear logging handlers!
    del logging.getLogger('press').handlers[:]
    # Clear hooks
    clear_hooks()


# noinspection PyBroadException
def entry(run_configuration, mercury_press_configuration):
    """

    :param run_configuration: The press state file
    :param mercury_press_configuration: press environment configuration from
    mercury-agent.yaml
    :return:
    """
    log.info('Initializing plugins')
    init_plugins(run_configuration,
                 mercury_press_configuration.get('plugins', {}).get(
                     'scan_directories'),
                 mercury_press_configuration.get('plugins', {}).get(
                     'enabled'
                 ))

    return_data = {}

    p = None
    try:
        p = PressOrchestrator(
            run_configuration,
            mercury_press_configuration.get('paths', {}).get(
                'parted', 'parted'),
            mercury_press_configuration.get('deployment_root', '/mnt/press'),
            mercury_press_configuration.get('staging_directory', '/.press'),
            mercury_press_configuration.get('layout', {}).get(
                'use_fibre_channel', False
            ),
            mercury_press_configuration.get('layout', {}).get(
                'loop_only', False
            ),
            mercury_press_configuration.get('partition_table', {}).get(
                'partition_start', 1048576
            ),
            mercury_press_configuration.get('partition_table', {}).get(
                'alignment', 1048576
            ),
            mercury_press_configuration.get('volume_group', {}).get(
                'pe_size', '4MiB'
            )
        )
    except Exception:
        exec_dict = parse_exception()
        log.error('Error during initialization: {}'.format(
            fancy_traceback_short(exec_dict)))
        return_data = {'error': True, 'message': 'Error during initialization',
                       'exception': exec_dict}
        cleanup_thread()

    if p:
        try:
            p.run()
        except Exception:
            exec_dict = parse_exception()
            log.error('Error during deployment: {}'.format(
                fancy_traceback_short(exec_dict)))
            return_data = {'error': True,
                           'message': 'Error during initialization',
                           'exception': exec_dict}
        finally:
            if p.layout.committed:
                time.sleep(2)
                p.teardown()

            cleanup_thread()

    return return_data


def add_mercury_plugin_data(press_configuration, task_id):
    temp_plugins = press_configuration.get('plugins', [])
    if 'mercury' not in temp_plugins:
        temp_plugins.append('mercury')
        press_configuration['plugins'] = temp_plugins

    press_configuration['mercury'] = {
        'task_id': task_id,
        'backend_zurl': get_configuration().agent.remote.backend_url
    }


@capability('press', description='Native press support in mercury', serial=True,
            kwarg_names=['configuration'], task_id_kwargs=True)
def press_native(**kwargs):
    press_configuration = kwargs['configuration']
    task_id = kwargs['task_id']

    mercury_configuration = get_configuration()
    add_mercury_plugin_data(press_configuration, task_id)

    backend_client = BackEndClient(
        mercury_configuration.agent.remote.backend_url)
    log.info('Starting press')
    start = time.time()
    backend_client.update_task({'task_id': task_id,
                                'action': 'Press: Launching'})
    return_data = entry(press_configuration,
                        mercury_configuration.get('press', {}))
    return_data['press_execution_time'] = time.time() - start
    return return_data



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import shlex
import subprocess


def get_name(interface):
    command = 'biosdevname --policy physical -i %s' % interface
    p = subprocess.Popen(shlex.split(command),
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()

    if p.returncode:
        if p.returncode == 2:
            # Interface name is not present, or biosdevname is not supported
            # for this interface
            return
        if p.returncode == 3:
            raise OSError('biosdevname must be run as the root user')
        if p.returncode == 4:
            # Running on a virtual machine
            return
        else:
            raise Exception('Problem running biosdevname: %d' % p.returncode)

    return out.strip()



from mercury.common.helpers import cli


class SimpleRACException(Exception):
    pass


class SimpleRAC(object):
    def __init__(self, path='racadm'):
        """

        :param path:
        """
        self.racadm_path = cli.find_in_path(path)
        if not self.racadm_path:
            raise SimpleRACException('Could not find racadm binary')

    def racadm(self, command):
        result = cli.run('sh {} {}'.format(self.racadm_path, command),
                         raise_exception=False, ignore_error=True)
        if result.returncode:
            raise SimpleRACException(
                'Error running racadm command: {}, code {}'.format(
                    command, result.returncode))

        return result.stdout

    @property
    def getsysinfo(self):
        """
        Simple 'parser' for getsysinfo command
        :return:
        """
        output = self.racadm('getsysinfo')
        sys_info = {}
        key = None
        for line in output.splitlines():
            if not line:
                continue
            if '=' not in line and line.strip()[-1] == ':':
                key = line[:-1]
                sys_info[key] = {}
                continue

            if key:
                sub_key, d = [term.strip() for term in line.split('=', 1)]
                if d == '::':  # :: seems to imply an empty field
                    d = None
                sys_info[key][sub_key] = d
        return sys_info



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mercury_agent.inspector.inspectors import inspector
from mercury_agent.inspector.hwlib.cpuinfo import CPUInfo


@inspector.expose('cpu')
def cpu_inspector():
    _cpu = []
    cpu_info = CPUInfo()

    processors = cpu_info.physical_index
    for _id in processors:
        processor = processors[_id][0]
        _proc_dict = dict()
        _proc_dict['physical_id'] = _id
        _proc_dict['cores'] = int(processor['cpu_cores'])
        _proc_dict['threads'] = int(processor['siblings'])
        _proc_dict['model_name'] = processor['model_name']
        _proc_dict['cache_size'] = processor['cache_size']
        _proc_dict['cache_alignment'] = int(processor['cache_alignment'])
        _proc_dict['flags'] = processor['flags'].split()
        # speed
        _proc_dict['frequency'] = CPUInfo.get_speed_info(processor)

        _cpu.append(_proc_dict)
        _cpu.sort(key=lambda k: k['physical_id'])

    return _cpu


if __name__ == '__main__':
    from pprint import pprint

    pprint(cpu_inspector())



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from setuptools import setup, find_packages


setup(
    name='mercury-agent',
    version='0.1.8',
    packages=find_packages(exclude=['tests']),
    url='http://www.mercurysoft.io',
    license='Apache-2.0',
    author='Jared Rodriguez',
    author_email='jared.rodriguez@rackspace.com',
    description='Mercury agent, hardware libraries, and inspectors',
    install_requires=[
        'mercury-core',
        'netifaces',
        'pyudev',
        'python-hpssa',
        'lxml',
        'press',
        'six',
        'msgpack-python',
        'pyzmq'
    ],
    entry_points="""
    [console_scripts]
    mercury-agent = mercury_agent.agent:main
    """,
)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from hpssa import HPSSA

from mercury_agent.configuration import get_configuration

from mercury_agent.hardware import platform_detection
from mercury_agent.hardware.drivers import driver, PCIDriverBase
from mercury_agent.hardware.raid.abstraction.api import RAIDActions, RAIDAbstractionException


class SmartArrayActions(RAIDActions):
    logical_drive_status_map = {
        'OK': 'OK',
        'Recovering': 'RECOVERING',
        'Interim Recovery Mode': 'DEGRADED',
        'Failed': 'FAILED'
    }

    physical_drive_status_map = {
        'OK': 'OK',
        'Rebuilding': 'REBUILDING',
        'Failed': 'FAILED'
    }

    def __init__(self):
        super(SmartArrayActions, self).__init__()
        self.hpssa = HPSSA(hpssa_path=get_configuration().get(
            'agent', {}).get(
            'hardware', {}).get(
            'raid', {}).get(
            'hpssacli_path') or 'hpssacli')

    @staticmethod
    def get_vendor_info(adapter):
        """
        :param adapter:
        :return: vendor_details
        """
        vendor_details = adapter.copy()

        # Configuration is transformed and top level, so remove the vendor configuration
        del vendor_details['configuration']

        # Drives can be found in the top level configuration
        del vendor_details['drives']

        # name is stored top level
        del vendor_details['name']

        return vendor_details

    def transform_logical_drives(self, logical_drives):
            return [{
                'status': self.logical_drive_status_map.get(logical_drive['status'], 'UNKNOWN'),
                'size': logical_drive['size'],
                'level': logical_drive['level'],
                'extra': {
                    'id': logical_drive['id'],
                    'progress': logical_drive['progress']
                }
            } for logical_drive in logical_drives]

    def convert_physical_drive(self, physical_drive):
        status = self.physical_drive_status_map.get(physical_drive['status'])
        if not status:
            raise RAIDAbstractionException(
                'Drive {port}:{box}:{bay} is in an unknown state: {status}'.format(
                    **physical_drive))

        new_physical_drive = {
            'size': physical_drive['size'],
            'status': status,
            'type': physical_drive['type'],
            'extra': {
                'port': physical_drive['port'],
                'box': physical_drive['box'],
                'bay': physical_drive['bay'],
            }
        }
        return new_physical_drive

    def transform_physical_drives(self, physical_drives):
        return [self.convert_physical_drive(physical_drive) for physical_drive in physical_drives]

    @staticmethod
    def get_array_index_from_letter(arrays, letter):
        for idx in range(len(arrays)):
            if arrays[idx]['extra']['letter'] == letter:
                return idx
        return -1

    def transform_configuration(self, original):
        configuration = {
            'arrays': [],
            'spares': [],
            'unassigned': []
        }

        for array in original['arrays']:
            configuration['arrays'].append({
                'free_space': array['free_space'],
                'extra': {
                    'letter': array['letter']
                },
                'logical_drives': self.transform_logical_drives(array['logical_drives']),
                'physical_drives': self.transform_physical_drives(array['physical_drives'])
            })

        #####################################################################################
        # hpssa spares look like this:
        #     [{'arrays': ['A'],
        #       'bay': '8',
        #       'box': '1',
        #       'port': '1I',
        #       'size': 300000000000,
        #       'status': 'OK',
        #       'type': 'SAS'}]
        # since this abstraction layer uses indexes for everything, we convert the letter
        # references to an index reference
        #####################################################################################

        for spare in original['spares']:
            array_indices = []
            for ref in spare['arrays']:
                idx = self.get_array_index_from_letter(configuration['arrays'], ref)
                if idx == -1:
                    raise RAIDAbstractionException('A spare is defined with an invalid array reference')
                array_indices.append(idx)

            spare_physical_drive = self.convert_physical_drive(spare)
            spare_physical_drive['target'] = array_indices
            configuration['spares'].append(spare_physical_drive)

        for unassigned in original['unassigned']:
            configuration['unassigned'].append(self.convert_physical_drive(unassigned))

        return configuration

    @staticmethod
    def get_slot(adapter_info):
        try:
            return int(adapter_info['vendor_info']['slot'])
        except KeyError:
            raise RAIDAbstractionException('Adapter is missing HP vendor/slot information')

    @staticmethod
    def get_letter_from_index(adapter_info, array_index):
        arrays = adapter_info['configuration']['arrays']

        try:
            our_array = arrays[array_index]
        except IndexError:
            raise RAIDAbstractionException('array {} does not exist'.format(array_index))

        # noinspection PyTypeChecker
        return our_array['extra']['letter']

    def transform_adapter_info(self, adapter_index):
        """
        Transforms python-hpssa adapter information into the standard format expected
        by RAIDActions
        :param adapter_index: list index of the adapter we are targeting
        :return: Adapter details in standard form
        """
        try:
            adapter = self.hpssa.adapters[adapter_index]
        except IndexError:
            raise RAIDAbstractionException('Adapter does not exist')

        adapter_details = {
            'name': adapter['name'],
            'provider': 'hpsa',
            'vendor_info': self.get_vendor_info(adapter),
            'configuration': self.transform_configuration(adapter['configuration'])
        }

        return adapter_details

    @staticmethod
    def sort_drives(drives):
        """
        :param drives:
        :return:
        """
        drives.sort(key=lambda x: '{}-{:05}-{:05}'.format(
            x['extra']['port'], int(x['extra']['box']),  int(x['extra']['bay'])))

    @staticmethod
    def assemble_drive(drive):
        return '{port}:{box}:{bay}'.format(**drive['extra'])

    def create(self, adapter_info, level=None, drives=None, size=None, array=None):
        """
        Implementation of RAIDActions.create

        Called from RAIDActions.create_logical_drive

        :param adapter_info: transformed adapter_info
        :param level: RAID level supported by RAIDActions API
        :param drives: Converted drive target from RAIDActions create_logical_drive
            or None if array is specified
        :type drives: list
        :param size: python-size Size() object or None
        :param array: If specified, update the array that matches the index
        :return: result (returncode, stdout, stderr)
        :return type: AttributeString
        """
        size_mb = size and size.megabytes or 'max'

        slot = self.get_slot(adapter_info)
        if level == '10':
            level = '1+0'

        if drives:
            hpsa_targets = []
            for drive in drives:
                hpsa_targets.append(self.assemble_drive(drive))

            result = self.hpssa.create(slot, selection=','.join(hpsa_targets), raid=level, size=size_mb)

        else:
            array_letter = array['extra']['letter']
            result = self.hpssa.create(slot, raid=level, array_letter=array_letter, size=size_mb)

        return result

    def delete_logical_drive(self, adapter, array, logical_drive):
        """
        Implementation for RAIDActions.delete_logical_drive

        :param adapter: adapter index
        :param array: array index
        :param logical_drive: logical drive index
        :return:
        """
        adapter_info = self.get_adapter_info(adapter)
        slot = self.get_slot(adapter_info)
        arrays = adapter_info['configuration']['arrays']
        try:
            target = arrays[array]['logical_drives'][logical_drive]
        except IndexError:
            raise RAIDAbstractionException(
                'Logical Drive does not exist at {}:{}:{}'.format(adapter, array, logical_drive))

        return self.hpssa.delete_logical_drive(slot, target['extra']['id'])

    def clear_configuration(self, adapter):
        """
        Implementation of RAIDActions.clear_configuration
        :param adapter: adapter index
        :return:
        """
        return self.hpssa.delete_all_logical_drives(
            self.get_slot(self.get_adapter_info(adapter)))

    def add_spares(self, adapter, drives, arrays=None):
        """
        Implementation of RAIDActions.add_spares

        :param adapter: adapter index
        :param arrays: array indexes
        :param drives: RAIDActions drive selection
        :return:
        """
        adapter_info = self.get_adapter_info(adapter)
        slot = self.get_slot(adapter_info)

        if arrays:
            array_letters = [self.get_letter_from_index(adapter_info, array)
                             for array in arrays]
        else:
            array_letters = None

        target_drives = self.get_drives_from_selection(adapter, drives)

        return self.hpssa.add_spares(slot,
                                     ','.join([self.assemble_drive(d)
                                               for d in target_drives]),
                                     array_letters)

    def erase(self, adapter, drives, method='fast'):
        raise NotImplementedError

    def get_erase_status(self, adapter):
        raise NotImplementedError

    def erase_in_progress(self, adapter):
        raise NotImplementedError


@driver()
class SmartArrayDriver(PCIDriverBase):
    name = 'hpssa'
    driver_type = 'raid'
    _handler = SmartArrayActions
    wants = 'pci'

    PCI_DEVICE_IDS = [
        "3239"  # Smart Array Gen9 Controllers
    ]

    @classmethod
    def probe(cls, pci_data):
        raid_pci_devices = platform_detection.get_raid_controllers(pci_data)
        if not platform_detection.has_smart_array_gen9(pci_data=raid_pci_devices):
            return

        owns = list()
        for device in raid_pci_devices:
            if cls.check(device):
                owns.append(device['slot'])
        return owns

    @classmethod
    def check(cls, pci_device):
        return pci_device['device_id'] in cls.PCI_DEVICE_IDS

    def inspect(self):
        adapters = []
        for idx in range(len(self.handler.hpssa.adapters)):
            _a = dict(**self.handler.get_adapter_info(idx))
            _a.update({
                'adapter_handler': self.name
            })

            adapters.append(_a)

        return adapters



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from size.size import Size


def parse_meminfo():
    with open('/proc/meminfo') as fp:
        data = fp.read()

    d = {}
    for line in data.splitlines():
        line = line.strip()
        if not line:
            continue
        k, v = line.split(':')
        v = v.strip()
        if 'kB'in v:
            v = Size(v).bytes
        d[k] = v

    return d

if __name__ == '__main__':
    print(parse_meminfo())
